"""
Engines are things that run around the reactor (see `kopf.reactor`)
to help it to function at full strength, but are not part of it.
For example, all never-ending side-tasks for peering and k8s-event-posting.

The reactor and engines exchange the state with each other (bi-directionally)
via the provided synchronization objects, usually asyncio events & queues.
"""



import pytest

import kopf
from kopf.reactor.handling import handler_var, subregistry_var
from kopf.reactor.invocation import context
from kopf.reactor.registries import OperatorRegistry, ResourceChangingRegistry
from kopf.structs.handlers import HANDLER_REASONS, Activity, ErrorsMode, Reason


def test_on_startup_minimal():
    registry = kopf.get_default_registry()

    @kopf.on.startup()
    def fn(**_):
        pass

    handlers = registry._activities.get_handlers(activity=Activity.STARTUP)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].activity == Activity.STARTUP
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None


def test_on_cleanup_minimal():
    registry = kopf.get_default_registry()

    @kopf.on.cleanup()
    def fn(**_):
        pass

    handlers = registry._activities.get_handlers(activity=Activity.CLEANUP)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].activity == Activity.CLEANUP
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None


def test_on_probe_minimal():
    registry = kopf.get_default_registry()

    @kopf.on.probe()
    def fn(**_):
        pass

    handlers = registry._activities.get_handlers(activity=Activity.PROBE)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].activity == Activity.PROBE
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None


# Resume handlers are mixed-in into all resource-changing reactions with initial listing.
@pytest.mark.parametrize('reason', HANDLER_REASONS)
def test_on_resume_minimal(reason, cause_factory, resource):
    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=reason, initial=True)

    @kopf.on.resume(*resource)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason is None
    assert handlers[0].field is None
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None
    assert handlers[0].when is None
    assert handlers[0].field is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_create_minimal(cause_factory, resource):
    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=Reason.CREATE)

    @kopf.on.create(*resource)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason == Reason.CREATE
    assert handlers[0].field is None
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None
    assert handlers[0].when is None
    assert handlers[0].field is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_update_minimal(cause_factory, resource):
    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=Reason.UPDATE)

    @kopf.on.update(*resource)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason == Reason.UPDATE
    assert handlers[0].field is None
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None
    assert handlers[0].when is None
    assert handlers[0].field is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_delete_minimal(cause_factory, resource):
    registry = kopf.get_default_registry()
    cause = cause_factory(resource=resource, reason=Reason.DELETE)

    @kopf.on.delete(*resource)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason == Reason.DELETE
    assert handlers[0].field is None
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None
    assert handlers[0].when is None
    assert handlers[0].field is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_field_minimal(cause_factory, resource):
    registry = kopf.get_default_registry()
    old = {'field': {'subfield': 'old'}}
    new = {'field': {'subfield': 'new'}}
    cause = cause_factory(resource=resource, reason=Reason.UPDATE, old=old, new=new, body=new)

    @kopf.on.field(*resource, field='field.subfield')
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason is None
    assert handlers[0].errors is None
    assert handlers[0].timeout is None
    assert handlers[0].retries is None
    assert handlers[0].backoff is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None
    assert handlers[0].when is None
    assert handlers[0].field == ('field', 'subfield')
    assert handlers[0].value is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_field_fails_without_field(resource):
    with pytest.raises(TypeError):
        @kopf.on.field(*resource)
        def fn(**_):
            pass


def test_on_startup_with_all_kwargs():
    registry = OperatorRegistry()

    @kopf.on.startup(
        id='id', registry=registry,
        errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78)
    def fn(**_):
        pass

    handlers = registry._activities.get_handlers(activity=Activity.STARTUP)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].activity == Activity.STARTUP
    assert handlers[0].id == 'id'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78


def test_on_cleanup_with_all_kwargs():
    registry = OperatorRegistry()

    @kopf.on.cleanup(
        id='id', registry=registry,
        errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78)
    def fn(**_):
        pass

    handlers = registry._activities.get_handlers(activity=Activity.CLEANUP)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].activity == Activity.CLEANUP
    assert handlers[0].id == 'id'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78


def test_on_probe_with_all_kwargs():
    registry = OperatorRegistry()

    @kopf.on.probe(
        id='id', registry=registry,
        errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78)
    def fn(**_):
        pass

    handlers = registry._activities.get_handlers(activity=Activity.PROBE)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].activity == Activity.PROBE
    assert handlers[0].id == 'id'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78


# Resume handlers are mixed-in into all resource-changing reactions with initial listing.
@pytest.mark.parametrize('reason', HANDLER_REASONS)
def test_on_resume_with_most_kwargs(mocker, reason, cause_factory, resource):
    registry = OperatorRegistry()
    cause = cause_factory(resource=resource, reason=reason, initial=True)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    when = lambda **_: False

    @kopf.on.resume(*resource,
                    id='id', registry=registry,
                    errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78,
                    deleted=True,
                    labels={'somelabel': 'somevalue'},
                    annotations={'someanno': 'somevalue'},
                    when=when)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason is None
    assert handlers[0].id == 'id'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78
    assert handlers[0].deleted == True
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}
    assert handlers[0].when == when
    assert handlers[0].field is None
    assert handlers[0].value is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_create_with_most_kwargs(mocker, cause_factory, resource):
    registry = OperatorRegistry()
    cause = cause_factory(resource=resource, reason=Reason.CREATE)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    when = lambda **_: False

    @kopf.on.create(*resource,
                    id='id', registry=registry,
                    errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78,
                    labels={'somelabel': 'somevalue'},
                    annotations={'someanno': 'somevalue'},
                    when=when)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason == Reason.CREATE
    assert handlers[0].id == 'id'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}
    assert handlers[0].when == when
    assert handlers[0].field is None
    assert handlers[0].value is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_update_with_most_kwargs(mocker, cause_factory, resource):
    registry = OperatorRegistry()
    cause = cause_factory(resource=resource, reason=Reason.UPDATE)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    when = lambda **_: False

    @kopf.on.update(*resource,
                    id='id', registry=registry,
                    errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78,
                    labels={'somelabel': 'somevalue'},
                    annotations={'someanno': 'somevalue'},
                    when=when)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason == Reason.UPDATE
    assert handlers[0].id == 'id'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}
    assert handlers[0].when == when
    assert handlers[0].field is None
    assert handlers[0].value is None
    assert handlers[0].old is None
    assert handlers[0].new is None


@pytest.mark.parametrize('optional', [
    pytest.param(True, id='optional'),
    pytest.param(False, id='mandatory'),
])
def test_on_delete_with_most_kwargs(mocker, cause_factory, optional, resource):
    registry = OperatorRegistry()
    cause = cause_factory(resource=resource, reason=Reason.DELETE)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    when = lambda **_: False

    @kopf.on.delete(*resource,
                    id='id', registry=registry,
                    errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78,
                    optional=optional,
                    labels={'somelabel': 'somevalue'},
                    annotations={'someanno': 'somevalue'},
                    when=when)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason == Reason.DELETE
    assert handlers[0].id == 'id'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}
    assert handlers[0].when == when
    assert handlers[0].field is None
    assert handlers[0].value is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_on_field_with_most_kwargs(mocker, cause_factory, resource):
    registry = OperatorRegistry()
    old = {'field': {'subfield': 'old'}}
    new = {'field': {'subfield': 'new'}}
    cause = cause_factory(resource=resource, reason=Reason.UPDATE, old=old, new=new, body=new)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    when = lambda **_: False

    @kopf.on.field(*resource, field='field.subfield',
                   id='id', registry=registry,
                   errors=ErrorsMode.PERMANENT, timeout=123, retries=456, backoff=78,
                   labels={'somelabel': 'somevalue'},
                   annotations={'someanno': 'somevalue'},
                   when=when)
    def fn(**_):
        pass

    handlers = registry._resource_changing.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].reason is None
    assert handlers[0].id == 'id/field.subfield'
    assert handlers[0].errors == ErrorsMode.PERMANENT
    assert handlers[0].timeout == 123
    assert handlers[0].retries == 456
    assert handlers[0].backoff == 78
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}
    assert handlers[0].when == when
    assert handlers[0].field == ('field', 'subfield')
    assert handlers[0].value is None
    assert handlers[0].old is None
    assert handlers[0].new is None


def test_subhandler_fails_with_no_parent_handler():

    registry = ResourceChangingRegistry()
    subregistry_var.set(registry)

    # Check if the contextvar is indeed not set (as a prerequisite).
    with pytest.raises(LookupError):
        handler_var.get()

    # Check the actual behaviour of the decorator.
    with pytest.raises(LookupError):
        @kopf.subhandler()
        def fn(**_):
            pass


def test_subhandler_declaratively(parent_handler, cause_factory):
    cause = cause_factory(reason=Reason.UPDATE)

    registry = ResourceChangingRegistry()
    subregistry_var.set(registry)

    with context([(handler_var, parent_handler)]):
        @kopf.subhandler()
        def fn(**_):
            pass

    handlers = registry.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn


def test_subhandler_imperatively(parent_handler, cause_factory):
    cause = cause_factory(reason=Reason.UPDATE)

    registry = ResourceChangingRegistry()
    subregistry_var.set(registry)

    def fn(**_):
        pass

    with context([(handler_var, parent_handler)]):
        kopf.register(fn)

    handlers = registry.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn


@pytest.mark.parametrize('decorator, kwargs', [
    (kopf.on.event, {}),
    (kopf.on.resume, {}),
    (kopf.on.create, {}),
    (kopf.on.update, {}),
    (kopf.on.delete, {}),
    (kopf.on.field, dict(field='x')),
])
def test_labels_filter_with_nones(resource, decorator, kwargs):

    with pytest.raises(ValueError):
        @decorator(*resource, **kwargs, labels={'x': None})
        def fn(**_):
            pass


@pytest.mark.parametrize('decorator, kwargs', [
    (kopf.on.event, {}),
    (kopf.on.resume, {}),
    (kopf.on.create, {}),
    (kopf.on.update, {}),
    (kopf.on.delete, {}),
    (kopf.on.field, dict(field='f')),
])
def test_annotations_filter_with_nones(resource, decorator, kwargs):

    with pytest.raises(ValueError):
        @decorator(*resource, **kwargs, annotations={'x': None})
        def fn(**_):
            pass


@pytest.mark.parametrize('decorator, causeargs, handlers_prop', [
    pytest.param(kopf.on.event, dict(), '_resource_watching', id='on-event'),
    pytest.param(kopf.on.resume, dict(reason=None, initial=True), '_resource_changing', id='on-resume'),
    pytest.param(kopf.on.create, dict(reason=Reason.CREATE), '_resource_changing', id='on-create'),
    pytest.param(kopf.on.update, dict(reason=Reason.UPDATE), '_resource_changing', id='on-update'),
    pytest.param(kopf.on.delete, dict(reason=Reason.DELETE), '_resource_changing', id='on-delete'),
    pytest.param(kopf.on.field, dict(reason=Reason.UPDATE), '_resource_changing', id='on-field'),
    pytest.param(kopf.daemon, dict(), '_resource_spawning', id='on-daemon'),
    pytest.param(kopf.timer, dict(), '_resource_spawning', id='on-timer'),
])
def test_field_with_value(mocker, cause_factory, decorator, causeargs, handlers_prop, resource, registry):
    old = {'field': {'subfield': 'old'}}
    new = {'field': {'subfield': 'new'}}
    cause = cause_factory(resource=resource, old=old, new=new, body=new, **causeargs)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    @decorator(*resource, field='spec.field', value='value')
    def fn(**_):
        pass

    handlers_registry = getattr(registry, handlers_prop)
    handlers = handlers_registry.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].field == ('spec', 'field')
    assert handlers[0].value == 'value'


@pytest.mark.parametrize('decorator, causeargs, handlers_prop', [
    pytest.param(kopf.on.update, dict(reason=Reason.UPDATE), '_resource_changing', id='on-update'),
    pytest.param(kopf.on.field, dict(reason=Reason.UPDATE), '_resource_changing', id='on-field'),
])
def test_field_with_oldnew(mocker, cause_factory, decorator, causeargs, handlers_prop, resource, registry):
    cause = cause_factory(resource=resource, **causeargs)
    mocker.patch('kopf.reactor.registries.match', return_value=True)

    @decorator(*resource, field='spec.field', old='old', new='new')
    def fn(**_):
        pass

    handlers_registry = getattr(registry, handlers_prop)
    handlers = handlers_registry.get_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].field == ('spec', 'field')
    assert handlers[0].value is None
    assert handlers[0].old == 'old'
    assert handlers[0].new == 'new'


@pytest.mark.parametrize('decorator', [
    pytest.param(kopf.on.event, id='on-event'),
    pytest.param(kopf.on.resume, id='on-resume'),
    pytest.param(kopf.on.create, id='on-create'),
    pytest.param(kopf.on.update, id='on-update'),
    pytest.param(kopf.on.delete, id='on-delete'),
    pytest.param(kopf.daemon, id='on-daemon'),
    pytest.param(kopf.timer, id='on-timer'),
])
def test_missing_field_with_specified_value(resource, decorator):
    with pytest.raises(TypeError, match="without a mandatory field"):
        @decorator(*resource, value='v')
        def fn(**_):
            pass


@pytest.mark.parametrize('kwargs', [
    pytest.param(dict(field='f', value='v', old='x'), id='value-vs-old'),
    pytest.param(dict(field='f', value='v', new='x'), id='value-vs-new'),
])
@pytest.mark.parametrize('decorator', [
    pytest.param(kopf.on.update, id='on-update'),
    pytest.param(kopf.on.field, id='on-field'),
])
def test_conflicts_of_values_vs_oldnew(resource, decorator, kwargs):
    with pytest.raises(TypeError, match="Either value= or old=/new="):
        @decorator(*resource, **kwargs)
        def fn(**_):
            pass


@pytest.mark.parametrize('decorator', [
    pytest.param(kopf.on.resume, id='on-resume'),
    pytest.param(kopf.on.create, id='on-create'),
    pytest.param(kopf.on.delete, id='on-delete'),
])
def test_invalid_oldnew_for_inappropriate_subhandlers(resource, decorator, registry):

    @decorator(*resource)
    def fn(**_):
        @kopf.subhandler(field='f', old='x')
        def fn2(**_):
            pass

    subregistry = ResourceChangingRegistry()
    handler = registry._resource_changing.get_all_handlers()[0]
    with context([(handler_var, handler), (subregistry_var, subregistry)]):
        with pytest.raises(TypeError, match="can only be used in update handlers"):
            handler.fn()



import asyncio

import pytest

from kopf.utilities.aiotasks import create_task


async def sample() -> None:
    pass


@pytest.mark.skipif('sys.version_info < (3, 8)')
def test_py38_create_task_is_the_native_one():
    assert create_task is asyncio.create_task


@pytest.mark.skipif('sys.version_info >= (3, 8)')
async def test_py37_create_task_accepts_name(mocker):
    real_create_task = mocker.patch('asyncio.create_task')
    coro = sample()
    task = create_task(coro, name='unused')
    assert real_create_task.called
    assert task is real_create_task.return_value
    await coro  # to prevent "never awaited" errors



"""
All the structures needed for Kubernetes patching.

Currently, it is implemented via a JSON merge-patch (RFC 7386),
i.e. a simple dictionary with field overrides, and ``None`` for field deletions.

In the future, it can be extended to a standalone object, which exposes
a dict-like behaviour, and remembers the changes in order of their execution,
and then generates the JSON patch (RFC 6902).
"""
from typing import Any, Dict, MutableMapping, Optional

from kopf.structs import dicts


class MetaPatch(dicts.MutableMappingView[str, Any]):
    _labels: dicts.MutableMappingView[str, Optional[str]]
    _annotations: dicts.MutableMappingView[str, Optional[str]]

    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'metadata')
        self._labels = dicts.MutableMappingView(self, 'labels')
        self._annotations = dicts.MutableMappingView(self, 'annotations')

    @property
    def labels(self) -> dicts.MutableMappingView[str, Optional[str]]:
        return self._labels

    @property
    def annotations(self) -> dicts.MutableMappingView[str, Optional[str]]:
        return self._annotations


class SpecPatch(dicts.MutableMappingView[str, Any]):
    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'spec')


class StatusPatch(dicts.MutableMappingView[str, Any]):
    def __init__(self, __src: "Patch") -> None:
        super().__init__(__src, 'status')


# Event-handling structures, used internally in the framework and handlers only.
class Patch(Dict[str, Any]):

    def __init__(self, __src: Optional[MutableMapping[str, Any]] = None) -> None:
        super().__init__(__src or {})
        self._meta = MetaPatch(self)
        self._spec = SpecPatch(self)
        self._status = StatusPatch(self)

    @property
    def metadata(self) -> MetaPatch:
        return self._meta

    @property
    def meta(self) -> MetaPatch:
        return self._meta

    @property
    def spec(self) -> SpecPatch:
        return self._spec

    @property
    def status(self) -> StatusPatch:
        return self._status



"""
A in-memory storage of arbitrary information per resource/object.

The information is stored strictly in-memory and is not persistent.
On the operator restart, all the memories are lost.

It is used internally to track allocated system resources for each Kubernetes
object, even if that object does not show up in the event streams for long time.
"""
import dataclasses
import logging
import time
from typing import Any, Dict, Iterator, MutableMapping, Optional, Set, Union

from kopf.structs import bodies, handlers, primitives
from kopf.utilities import aiotasks


@dataclasses.dataclass(frozen=True)
class Daemon:
    task: aiotasks.Task  # a guarding task of the daemon.
    logger: Union[logging.Logger, logging.LoggerAdapter]
    handler: handlers.ResourceSpawningHandler
    stopper: primitives.DaemonStopper  # a signaller for the termination and its reason.


@dataclasses.dataclass(frozen=False)
class Throttler:
    """ A state of throttling for one specific purpose (there can be a few). """
    source_of_delays: Optional[Iterator[float]] = None
    last_used_delay: Optional[float] = None
    active_until: Optional[float] = None  # internal clock


class Memo(Dict[Any, Any]):
    """ A container to hold arbitrary keys-fields assigned by the users. """

    def __setattr__(self, key: str, value: Any) -> None:
        self[key] = value

    def __delattr__(self, key: str) -> None:
        try:
            del self[key]
        except KeyError as e:
            raise AttributeError(str(e))

    def __getattr__(self, key: str) -> Any:
        try:
            return self[key]
        except KeyError as e:
            raise AttributeError(str(e))


@dataclasses.dataclass(frozen=False)
class ResourceMemory:
    """ A system memo about a single resource/object. Usually stored in `Memories`. """

    # For arbitrary user data to be stored in memory, passed as `memo` to all the handlers.
    memo: Memo = dataclasses.field(default_factory=Memo)

    # For resuming handlers tracking and deciding on should they be called or not.
    noticed_by_listing: bool = False
    fully_handled_once: bool = False

    # Throttling for API errors (mostly from PATCHing) and for processing in general.
    error_throttler: Throttler = dataclasses.field(default_factory=Throttler)

    # For background and timed threads/tasks (invoked with the kwargs of the last-seen body).
    live_fresh_body: Optional[bodies.Body] = None
    idle_reset_time: float = dataclasses.field(default_factory=time.monotonic)
    forever_stopped: Set[handlers.HandlerId] = dataclasses.field(default_factory=set)
    running_daemons: Dict[handlers.HandlerId, Daemon] = dataclasses.field(default_factory=dict)


class ResourceMemories:
    """
    A container of all memos about every existing resource in a single operator.

    Distinct operator tasks have their own memory containers, which
    do not overlap. This solves the problem if storing the per-resource
    entries in the global or context variables.

    The memos can store anything the resource handlers need to persist within
    a single process/operator lifetime, but not persisted on the resource.
    For example, the runtime system resources: flags, threads, tasks, etc.
    Or the scalar values, which have meaning only for this operator process.

    The container is relatively async-safe: one individual resource is always
    handled sequentially, never in parallel with itself (different resources
    are handled in parallel through), so the same key will not be added/deleted
    in the background during the operation, so the locking is not needed.
    """
    _items: MutableMapping[str, ResourceMemory]

    def __init__(self) -> None:
        super().__init__()
        self._items = {}

    def iter_all_memories(self) -> Iterator[ResourceMemory]:
        for memory in self._items.values():
            yield memory

    async def recall(
            self,
            raw_body: bodies.RawBody,
            *,
            noticed_by_listing: bool = False,
    ) -> ResourceMemory:
        """
        Either find a resource's memory, or create and remember a new one.

        Keep the last-seen body up to date for all the handlers.
        """
        key = self._build_key(raw_body)
        if key not in self._items:
            memory = ResourceMemory(noticed_by_listing=noticed_by_listing)
            self._items[key] = memory
        return self._items[key]

    async def forget(
            self,
            raw_body: bodies.RawBody,
    ) -> None:
        """
        Forget the resource's memory if it exists; or ignore if it does not.
        """
        key = self._build_key(raw_body)
        if key in self._items:
            del self._items[key]

    def _build_key(
            self,
            raw_body: bodies.RawBody,
    ) -> str:
        """
        Construct an immutable persistent key of a resource.

        Generally, a uid is sufficient, as it is unique within the cluster.
        But it can be e.g. plural/namespace/name triplet, or anything else,
        even of different types (as long as it satisfies the type checkers).

        But it must be consistent within a single process lifetime.
        """
        return raw_body.get('metadata', {}).get('uid') or ''



import asyncio
import dataclasses
from unittest.mock import Mock

import freezegun
import pytest

from kopf.engines.peering import process_peering_event
from kopf.structs import bodies, primitives


@dataclasses.dataclass(frozen=True, eq=False)
class K8sMocks:
    patch_obj: Mock


@pytest.fixture(autouse=True)
def k8s_mocked(mocker, resp_mocker):
    # We mock on the level of our own K8s API wrappers, not the K8s client.
    return K8sMocks(
        patch_obj=mocker.patch('kopf.clients.patching.patch_obj', return_value={}),
    )


@pytest.fixture
async def replenished(mocker):
    # Make sure that freeze-sleeps are not actually executed, i.e. exit instantly.
    replenished = asyncio.Event()
    replenished.set()
    mocker.patch.object(replenished, 'wait')  # to avoid RuntimeWarnings for unwaited coroutines
    return replenished


async def test_other_peering_objects_are_ignored(
        mocker, k8s_mocked, settings, replenished,
        peering_resource, peering_namespace):

    status = mocker.Mock()
    status.items.side_effect = Exception("This should not be called.")
    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'their-name'},
            'status': status,
        })

    wait_for = mocker.patch('asyncio.wait_for')

    settings.peering.name = 'our-name'
    await process_peering_event(
        raw_event=event,
        freeze_toggle=primitives.Toggle(),
        replenished=replenished,
        autoclean=False,
        identity='id',
        settings=settings,
        namespace=peering_namespace,
        resource=peering_resource,
    )
    assert not status.items.called
    assert not k8s_mocked.patch_obj.called
    assert wait_for.call_count == 0


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_on_for_higher_priority_peer_when_initially_off(
        mocker, k8s_mocked, replenished, caplog, assert_logs, settings,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 101,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    freeze_toggle = primitives.Toggle(False)
    wait_for = mocker.patch('asyncio.wait_for')

    caplog.set_level(0)
    assert freeze_toggle.is_off()
    await process_peering_event(
        raw_event=event,
        freeze_toggle=freeze_toggle,
        replenished=replenished,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert freeze_toggle.is_on()
    assert wait_for.call_count == 1
    assert 9 < wait_for.call_args[1]['timeout'] < 10
    assert not k8s_mocked.patch_obj.called
    assert_logs(["Freezing operations in favour of"], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_higher_priority_peer_when_already_on(
        mocker, k8s_mocked, replenished, caplog, assert_logs, settings,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 101,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    freeze_toggle = primitives.Toggle(True)
    wait_for = mocker.patch('asyncio.wait_for')

    caplog.set_level(0)
    assert freeze_toggle.is_on()
    await process_peering_event(
        raw_event=event,
        freeze_toggle=freeze_toggle,
        replenished=replenished,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert freeze_toggle.is_on()
    assert wait_for.call_count == 1
    assert 9 < wait_for.call_args[1]['timeout'] < 10
    assert not k8s_mocked.patch_obj.called
    assert_logs([], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_off_for_lower_priority_peer_when_initially_on(
        mocker, k8s_mocked, replenished, caplog, assert_logs, settings,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 99,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    freeze_toggle = primitives.Toggle(True)
    wait_for = mocker.patch('asyncio.wait_for')

    caplog.set_level(0)
    assert freeze_toggle.is_on()
    await process_peering_event(
        raw_event=event,
        freeze_toggle=freeze_toggle,
        replenished=replenished,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert freeze_toggle.is_off()
    assert wait_for.call_count == 0
    assert not k8s_mocked.patch_obj.called
    assert_logs(["Resuming operations after the freeze"], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Freezing operations in favour of",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_lower_priority_peer_when_already_off(
        mocker, k8s_mocked, replenished, caplog, assert_logs, settings,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 99,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    freeze_toggle = primitives.Toggle(False)
    wait_for = mocker.patch('asyncio.wait_for')

    caplog.set_level(0)
    assert freeze_toggle.is_off()
    await process_peering_event(
        raw_event=event,
        freeze_toggle=freeze_toggle,
        replenished=replenished,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert freeze_toggle.is_off()
    assert wait_for.call_count == 0
    assert not k8s_mocked.patch_obj.called
    assert_logs([], prohibited=[
        "Possibly conflicting operators",
        "Freezing all operators, including self",
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_toggled_on_for_same_priority_peer_when_initially_off(
        mocker, k8s_mocked, replenished, caplog, assert_logs, settings,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 100,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    freeze_toggle = primitives.Toggle(False)
    wait_for = mocker.patch('asyncio.wait_for')

    caplog.set_level(0)
    assert freeze_toggle.is_off()
    await process_peering_event(
        raw_event=event,
        freeze_toggle=freeze_toggle,
        replenished=replenished,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert freeze_toggle.is_on()
    assert wait_for.call_count == 1
    assert 9 < wait_for.call_args[1]['timeout'] < 10
    assert not k8s_mocked.patch_obj.called
    assert_logs([
        "Possibly conflicting operators",
        "Freezing all operators, including self",
    ], prohibited=[
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
async def test_ignored_for_same_priority_peer_when_already_on(
        mocker, k8s_mocked, replenished, caplog, assert_logs, settings,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': 100,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    freeze_toggle = primitives.Toggle(True)
    wait_for = mocker.patch('asyncio.wait_for')

    caplog.set_level(0)
    assert freeze_toggle.is_on()
    await process_peering_event(
        raw_event=event,
        freeze_toggle=freeze_toggle,
        replenished=replenished,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert freeze_toggle.is_on()
    assert wait_for.call_count == 1
    assert 9 < wait_for.call_args[1]['timeout'] < 10
    assert not k8s_mocked.patch_obj.called
    assert_logs([
        "Possibly conflicting operators",
    ], prohibited=[
        "Freezing all operators, including self",
        "Freezing operations in favour of",
        "Resuming operations after the freeze",
    ])


@freezegun.freeze_time('2020-12-31T23:59:59.123456')
@pytest.mark.parametrize('priority', [100, 101])
async def test_resumes_immediately_on_expiration_of_blocking_peers(
        mocker, k8s_mocked, replenished, caplog, assert_logs, settings, priority,
        peering_resource, peering_namespace):

    event = bodies.RawEvent(
        type='ADDED',  # irrelevant
        object={
            'metadata': {'name': 'name', 'namespace': peering_namespace},  # for matching
            'status': {
                'higher-prio': {
                    'priority': priority,
                    'lifetime': 10,
                    'lastseen': '2020-12-31T23:59:59'
                },
            },
        })
    settings.peering.name = 'name'
    settings.peering.priority = 100

    freeze_toggle = primitives.Toggle(True)
    wait_for = mocker.patch('asyncio.wait_for', side_effect=asyncio.TimeoutError)

    caplog.set_level(0)
    assert freeze_toggle.is_on()
    await process_peering_event(
        raw_event=event,
        freeze_toggle=freeze_toggle,
        replenished=replenished,
        autoclean=False,
        namespace=peering_namespace,
        resource=peering_resource,
        identity='id',
        settings=settings,
    )
    assert freeze_toggle.is_on()
    assert wait_for.call_count == 1
    assert 9 < wait_for.call_args[1]['timeout'] < 10
    assert k8s_mocked.patch_obj.called



import asyncio
import time

import kopf


# Sync daemons in threads are non-interruptable, they must check for the `stopped` flag.
# This daemon exits after 3 attempts and then 30 seconds of running (unless stopped).
@kopf.daemon('kopfexamples', backoff=3)
def background_sync(spec, stopped, logger, retry, patch, **_):
    if retry < 3:
        patch.status['message'] = f"Failed {retry+1} times."
        raise kopf.TemporaryError("Simulated failure.", delay=1)

    started = time.time()
    while not stopped and time.time() - started <= 30:
        logger.info(f"=> Ping from a sync daemon: field={spec['field']!r}, retry={retry}")
        stopped.wait(5.0)

    patch.status['message'] = "Accompanying is finished."


# Async daemons do not need the `stopped` signal, they can rely on `asyncio.CancelledError` raised.
# This daemon runs forever (until stopped, i.e. cancelled). Yet fails to start for 3 first times.
@kopf.daemon('kopfexamples', backoff=3, cancellation_backoff=1.0, cancellation_timeout=0.5,
             annotations={'someannotation': 'somevalue'})
async def background_async(spec, logger, retry, **_):
    if retry < 3:
        raise kopf.TemporaryError("Simulated failure.", delay=1)

    while True:
        logger.info(f"=> Ping from an async daemon: field={spec['field']!r}")
        await asyncio.sleep(5.0)


E2E_CREATION_STOP_WORDS = ["=> Ping from"]
E2E_DELETION_STOP_WORDS = ["'background_async' is cancelled", "'background_sync' is cancelled", "'background_async' has exited"]



import aiohttp.web
import pytest

from kopf.clients.errors import APIError
from kopf.clients.scanning import scan_resources


async def test_no_resources_in_empty_apis(
        resp_mocker, aresponses, hostname):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': []}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': []}))

    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)

    resources = await scan_resources()
    assert len(resources) == 0

    assert core_mock.call_count == 1
    assert apis_mock.call_count == 1


@pytest.mark.parametrize('namespaced', [True, False])
async def test_resources_in_old_apis(
        resp_mocker, aresponses, hostname, namespaced):


    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': ['v1']}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': []}))
    scan_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': [
        {
            'kind': 'kind1',
            'name': 'plural1',
            'singularName': 'singular1',
            'namespaced': namespaced,
            'categories': ['category1', 'category2'],
            'shortNames': ['shortname1', 'shortname2'],
            'verbs': ['verb1', 'verb2'],
        },
    ]}))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/api/v1', 'get', scan_mock)

    resources = await scan_resources()
    assert len(resources) == 1

    resource1 = list(resources)[0]
    assert resource1.group == ''
    assert resource1.version == 'v1'
    assert resource1.kind == 'kind1'
    assert resource1.plural == 'plural1'
    assert resource1.singular == 'singular1'
    assert resource1.preferred == True
    assert resource1.namespaced == namespaced
    assert resource1.subresources == set()
    assert resource1.categories == {'category1', 'category2'}
    assert resource1.shortcuts == {'shortname1', 'shortname2'}
    assert resource1.verbs == {'verb1', 'verb2'}

    assert core_mock.call_count == 1
    assert apis_mock.call_count == 1
    assert scan_mock.call_count == 1


@pytest.mark.parametrize('namespaced', [True, False])
@pytest.mark.parametrize('preferred_version, expected_preferred', [
    ('version1', True),
    ('versionX', False),
])
async def test_resources_in_new_apis(
        resp_mocker, aresponses, hostname, namespaced,
        preferred_version, expected_preferred):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': []}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': [
        {
            'name': 'group1',
            'preferredVersion': {'version': preferred_version},
            'versions': [{'version': 'version1'}],
        },
    ]}))
    g1v1_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': [
        {
            'kind': 'kind1',
            'name': 'plural1',
            'singularName': 'singular1',
            'namespaced': namespaced,
            'categories': ['category1', 'category2'],
            'shortNames': ['shortname1', 'shortname2'],
            'verbs': ['verb1', 'verb2'],
        },
    ]}))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/apis/group1/version1', 'get', g1v1_mock)

    resources = await scan_resources()
    assert len(resources) == 1

    resource1 = list(resources)[0]
    assert resource1.group == 'group1'
    assert resource1.version == 'version1'
    assert resource1.kind == 'kind1'
    assert resource1.plural == 'plural1'
    assert resource1.singular == 'singular1'
    assert resource1.preferred == expected_preferred
    assert resource1.namespaced == namespaced
    assert resource1.subresources == set()
    assert resource1.categories == {'category1', 'category2'}
    assert resource1.shortcuts == {'shortname1', 'shortname2'}
    assert resource1.verbs == {'verb1', 'verb2'}

    assert core_mock.call_count == 1
    assert apis_mock.call_count == 1
    assert g1v1_mock.call_count == 1


async def test_subresources_in_old_apis(
        resp_mocker, aresponses, hostname):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': ['v1']}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': []}))
    v1v1_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': [
        {
            'kind': 'kind1',
            'name': 'plural1',
            'singularName': 'singular1',
            'namespaced': True,
            'categories': [],
            'shortNames': [],
            'verbs': [],
        },
        {
            'name': 'plural1/sub1',
        },
        {
            'name': 'plural1/sub2',
        },
        {
            'name': 'pluralX/sub3',
        },
    ]}))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/api/v1', 'get', v1v1_mock)

    resources = await scan_resources()
    assert len(resources) == 1
    resource1 = list(resources)[0]
    assert resource1.subresources == {'sub1', 'sub2'}


async def test_subresources_in_new_apis(
        resp_mocker, aresponses, hostname):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': []}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': [
        {
            'name': 'group1',
            'preferredVersion': {'version': 'version1'},
            'versions': [{'version': 'version1'}],
        },
    ]}))
    g1v1_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': [
        {
            'kind': 'kind1',
            'name': 'plural1',
            'singularName': 'singular1',
            'namespaced': True,
            'categories': [],
            'shortNames': [],
            'verbs': [],
        },
        {
            'name': 'plural1/sub1',
        },
        {
            'name': 'plural1/sub2',
        },
        {
            'name': 'pluralX/sub3',
        },
    ]}))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/apis/group1/version1', 'get', g1v1_mock)

    resources = await scan_resources()
    assert len(resources) == 1
    resource1 = list(resources)[0]
    assert resource1.subresources == {'sub1', 'sub2'}



@pytest.mark.parametrize('group_filter, exp_core, exp_apis, exp_crv1, exp_g1v1, exp_g2v1', [
    pytest.param([''], 1, 0, 1, 0, 0, id='only-core'),
    pytest.param(['g1'], 0, 1, 0, 1, 0, id='only-g1'),
    pytest.param(['g2'], 0, 1, 0, 0, 1, id='only-g2'),
    pytest.param(['', 'g1'], 1, 1, 1, 1, 0, id='core-and-g1'),
    pytest.param(['', 'g2'], 1, 1, 1, 0, 1, id='core-and-g2'),
    pytest.param(['g1', 'g2'], 0, 1, 0, 1, 1, id='g1-and-g2'),
    pytest.param(['X'], 0, 1, 0, 0, 0, id='unexistent'),
    pytest.param([], 0, 0, 0, 0, 0, id='restrictive'),
    pytest.param(None, 1, 1, 1, 1, 1, id='unfiltered'),
])
async def test_group_filtering(
        resp_mocker, aresponses, hostname,
        group_filter, exp_core, exp_apis, exp_crv1, exp_g1v1, exp_g2v1):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': ['v1']}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': [
        {'name': 'g1', 'preferredVersion': {'version': ''}, 'versions': [{'version': 'g1v1'}]},
        {'name': 'g2', 'preferredVersion': {'version': ''}, 'versions': [{'version': 'g2v1'}]},
    ]}))
    crv1_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': []}))
    g1v1_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': []}))
    g2v1_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': []}))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/api/v1', 'get', crv1_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/apis/g1/g1v1', 'get', g1v1_mock)
    aresponses.add(hostname, '/apis/g2/g2v1', 'get', g2v1_mock)

    await scan_resources(groups=group_filter)

    assert core_mock.call_count == exp_core
    assert apis_mock.call_count == exp_apis

    assert crv1_mock.call_count == exp_crv1
    assert g1v1_mock.call_count == exp_g1v1
    assert g2v1_mock.call_count == exp_g2v1



@pytest.mark.parametrize('status', [404])
async def test_http404_returns_no_resources_from_old_apis(
        resp_mocker, aresponses, hostname, status):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': ['v1']}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': []}))
    status_mock = resp_mocker(return_value=aresponses.Response(status=status))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/api/v1', 'get', status_mock)

    resources = await scan_resources()

    assert not resources
    assert status_mock.call_count == 1


@pytest.mark.parametrize('status', [404])
async def test_http404_returns_no_resources_from_new_apis(
        resp_mocker, aresponses, hostname, status):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': []}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': [
        {'name': 'g1', 'preferredVersion': {'version': ''}, 'versions': [{'version': 'g1v1'}]},
    ]}))
    status_mock = resp_mocker(return_value=aresponses.Response(status=status))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/apis/g1/g1v1', 'get', status_mock)

    resources = await scan_resources()

    assert not resources
    assert status_mock.call_count == 1


@pytest.mark.parametrize('status', [403, 500, 666])
async def test_unknown_api_statuses_escalate_from_old_apis(
        resp_mocker, aresponses, hostname, status):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': ['v1']}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': []}))
    status_mock = resp_mocker(return_value=aresponses.Response(status=status))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/api/v1', 'get', status_mock)

    with pytest.raises(APIError) as err:
        await scan_resources()

    assert err.value.status == status
    assert status_mock.call_count == 1


@pytest.mark.parametrize('status', [403, 500, 666])
async def test_unknown_api_statuses_escalate_from_new_apis(
        resp_mocker, aresponses, hostname, status):

    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': []}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': [
        {'name': 'g1', 'preferredVersion': {'version': ''}, 'versions': [{'version': 'g1v1'}]},
    ]}))
    status_mock = resp_mocker(return_value=aresponses.Response(status=status))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/apis/g1/g1v1', 'get', status_mock)

    with pytest.raises(APIError) as err:
        await scan_resources()

    assert err.value.status == status
    assert status_mock.call_count == 1


async def test_empty_singulars_fall_back_to_kinds(
        resp_mocker, aresponses, hostname):

    # Only one endpoint is enough, core v1 is easier to mock:
    core_mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': ['v1']}))
    apis_mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': []}))
    scan_mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': [
        {
            'kind': 'MultiWordKind',
            'name': '...',
            'singularName': '',  # as in K3s
            'namespaced': True,
            'categories': [],
            'shortNames': [],
            'verbs': [],
        },
    ]}))
    aresponses.add(hostname, '/api', 'get', core_mock)
    aresponses.add(hostname, '/apis', 'get', apis_mock)
    aresponses.add(hostname, '/api/v1', 'get', scan_mock)

    resources = await scan_resources(groups=[''])
    assert len(resources) == 1

    resource1 = list(resources)[0]
    assert resource1.singular == 'multiwordkind'


# TODO: LATER: test that the requests are done in parallel, and the total timing is the best possible.



from unittest.mock import Mock

from kopf.reactor.registries import _matches_resource
from kopf.structs.references import Resource, Selector


def test_different_resource():
    selector = Selector('group1', 'version1', 'plural1')
    resource = Resource('group2', 'version2', 'plural2')
    handler = Mock(selector=selector)
    matches = _matches_resource(handler, resource)
    assert not matches


def test_equivalent_resources():
    selector = Selector('group1', 'version1', 'plural1')
    resource = Resource('group1', 'version1', 'plural1')
    handler = Mock(selector=selector)
    matches = _matches_resource(handler, resource)
    assert matches


def test_catchall_with_none():
    resource = Resource('group2', 'version2', 'plural2')
    handler = Mock(selector=None)
    matches = _matches_resource(handler, resource)
    assert matches



import asyncio
import dataclasses
import importlib
import io
import json
import logging
import re
import sys
import time
from unittest.mock import Mock

import aiohttp.web
import asynctest
import pytest
import pytest_mock

import kopf
from kopf.clients.auth import APIContext
from kopf.engines.loggers import ObjectPrefixingTextFormatter, configure
from kopf.engines.posting import settings_var
from kopf.reactor.registries import OperatorRegistry
from kopf.structs.configuration import OperatorSettings
from kopf.structs.credentials import ConnectionInfo, Vault, VaultKey
from kopf.structs.references import Resource, Selector


def pytest_configure(config):
    config.addinivalue_line('markers', "e2e: end-to-end tests with real operators.")

    # Unexpected warnings should fail the tests. Use `-Wignore` to explicitly disable it.
    config.addinivalue_line('filterwarnings', 'error')

    # Warnings from the testing tools out of our control should not fail the tests.
    config.addinivalue_line('filterwarnings', 'ignore:"@coroutine":DeprecationWarning:asynctest.mock')
    config.addinivalue_line('filterwarnings', 'ignore:The loop argument:DeprecationWarning:aiohttp')
    config.addinivalue_line('filterwarnings', 'ignore:The loop argument:DeprecationWarning:aiojobs')
    config.addinivalue_line('filterwarnings', 'ignore:The loop argument:DeprecationWarning:asyncio.queues')  # aiojobs


def pytest_addoption(parser):
    parser.addoption("--only-e2e", action="store_true", help="Execute end-to-end tests only.")
    parser.addoption("--with-e2e", action="store_true", help="Include end-to-end tests.")


# Make all tests in this directory and below asyncio-compatible by default.
# Due to how pytest-async checks for these markers, they should be added as early as possible.
@pytest.hookimpl(hookwrapper=True)
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name) and asyncio.iscoroutinefunction(obj):
        pytest.mark.asyncio(obj)
    yield


# This logic is not applied if pytest is started explicitly on ./examples/.
# In that case, regular pytest behaviour applies -- this is intended.
def pytest_collection_modifyitems(config, items):

    # Put all e2e tests to the end, as they are assumed to be slow.
    def _is_e2e(item):
        path = item.location[0]
        return path.startswith('tests/e2e/') or path.startswith('examples/')
    etc = [item for item in items if not _is_e2e(item)]
    e2e = [item for item in items if _is_e2e(item)]

    # Mark all e2e tests, no matter how they were detected. Just for filtering.
    mark_e2e = pytest.mark.e2e
    for item in e2e:
        item.add_marker(mark_e2e)

    # Minikube tests are heavy and require a cluster. Skip them by default,
    # so that the contributors can run pytest without initial tweaks.
    mark_skip = pytest.mark.skip(reason="E2E tests are not enabled. "
                                        "Use --with-e2e/--only-e2e to enable.")
    if not config.getoption('--with-e2e') and not config.getoption('--only-e2e'):
        for item in e2e:
            item.add_marker(mark_skip)

    # Minify the test-plan if only e2e are requested (all other should be skipped).
    if config.getoption('--only-e2e'):
        items[:] = e2e
    else:
        items[:] = etc + e2e


# Substitute the regular mock with the async-aware mock in the `mocker` fixture.
@pytest.fixture(scope='session', autouse=True)
def enforce_asyncio_mocker(pytestconfig):
    pytest_mock.plugin.get_mock_module = lambda config: asynctest
    pytest_mock.get_mock_module = pytest_mock.plugin.get_mock_module
    fixture = pytest_mock.MockerFixture(pytestconfig)
    assert fixture.mock_module is asynctest, "Mock replacement failed!"


@pytest.fixture(params=[
    ('kopf.dev', 'v1', 'clusterkopfpeerings'),
    ('zalando.org', 'v1', 'clusterkopfpeerings'),
], ids=['kopf-dev-namespaced', 'zalando-org-namespaced'])
def namespaced_peering_resource(request):
    return Resource(*request.param[:3], namespaced=True)


@pytest.fixture(params=[
    ('kopf.dev', 'v1', 'kopfpeerings'),
    ('zalando.org', 'v1', 'kopfpeerings'),
], ids=['kopf-dev-cluster', 'zalando-org-cluster'])
def cluster_peering_resource(request):
    return Resource(*request.param[:3], namespaced=False)


@pytest.fixture(params=[
    ('kopf.dev', 'v1', 'clusterkopfpeerings', False),
    ('zalando.org', 'v1', 'clusterkopfpeerings', False),
    ('kopf.dev', 'v1', 'kopfpeerings', True),
    ('zalando.org', 'v1', 'kopfpeerings', True),
], ids=['kopf-dev-cluster', 'zalando-org-cluster', 'kopf-dev-namespaced', 'zalando-org-namespaced'])
def peering_resource(request):
    return Resource(*request.param[:3], namespaced=request.param[3])


@pytest.fixture()
def namespaced_resource():
    """ The resource used in the tests. Usually mocked, so it does not matter. """
    return Resource('kopf.dev', 'v1', 'kopfexamples', namespaced=True)


@pytest.fixture()
def cluster_resource():
    """ The resource used in the tests. Usually mocked, so it does not matter. """
    return Resource('kopf.dev', 'v1', 'kopfexamples', namespaced=False)


@pytest.fixture(params=[True, False], ids=['namespaced', 'cluster'])
def resource(request):
    """ The resource used in the tests. Usually mocked, so it does not matter. """
    return Resource('kopf.dev', 'v1', 'kopfexamples', namespaced=request.param)


@pytest.fixture()
def selector(resource):
    """ The selector used in the tests. Usually mocked, so it does not matter. """
    return Selector(group=resource.group, version=resource.version, plural=resource.plural)


@pytest.fixture()
def peering_namespace(peering_resource):
    return 'ns' if peering_resource.namespaced else None


@pytest.fixture()
def namespace(resource):
    return 'ns' if resource.namespaced else None


@pytest.fixture()
def settings():
    return OperatorSettings()


@pytest.fixture()
def settings_via_contextvar(settings):
    token = settings_var.set(settings)
    try:
        yield
    finally:
        settings_var.reset(token)


#
# Mocks for Kopf's internal but global variables.
#


@pytest.fixture
def registry_factory():
    # For most tests: not SmartOperatorRegistry, but the empty one!
    # For e2e tests: overridden to SmartOperatorRegistry.
    return OperatorRegistry


@pytest.fixture(autouse=True)
def registry(registry_factory):
    """
    Ensure that the tests have a fresh new global (not re-used) registry.
    """
    old_registry = kopf.get_default_registry()
    new_registry = registry_factory()
    kopf.set_default_registry(new_registry)
    yield new_registry
    kopf.set_default_registry(old_registry)


#
# Mocks for Kubernetes API clients (any of them). Reasons:
# 1. We do not test the clients, we test the layers on top of them,
#    so everything low-level should be mocked and assumed to be functional.
# 2. No external calls must be made under any circumstances.
#    The unit-tests must be fully isolated from the environment.
#

@pytest.fixture()
async def enforced_context(fake_vault, mocker):
    """
    Patchable context/session for some tests, e.g. with local exceptions.

    The local exceptions are supposed to simulate either the code issues,
    or the connection issues. `aresponses` does not allow to raise arbitrary
    exceptions on the client side, but only to return the erroneous responses.

    This test forces the re-authenticating decorators to always use one specific
    session for the duration of the test, so that the patches would have effect.
    """
    _, item = fake_vault.select()
    context = APIContext(item.info)
    mocker.patch(f'{APIContext.__module__}.{APIContext.__name__}', return_value=context)
    async with context.session:
        yield context


@pytest.fixture()
async def enforced_session(enforced_context: APIContext):
    yield enforced_context.session


# Note: Unused `fake_vault` is to ensure that the client wrappers have the credentials.
# Note: Unused `enforced_session` is to ensure that the session is closed for every test.
@pytest.fixture()
def resp_mocker(fake_vault, enforced_session, aresponses):
    """
    A factory of server-side callbacks for `aresponses` with mocking/spying.

    The value of the fixture is a function, which return a coroutine mock.
    That coroutine mock should be passed to `aresponses.add` as a response
    callback function. When called, it calls the mock defined by the function's
    arguments (specifically, return_value or side_effects).

    The difference from passing the responses directly to `aresponses.add`
    is that it is possible to assert on whether the response was handled
    by that callback at all (i.e. HTTP URL & method matched), especially
    if there are multiple responses registered.

    Sample usage::

        def test_me(resp_mocker):
            response = aiohttp.web.json_response({'a': 'b'})
            callback = resp_mocker(return_value=response)
            aresponses.add(hostname, '/path/', 'get', callback)
            do_something()
            assert callback.called
            assert callback.call_count == 1
    """
    def resp_maker(*args, **kwargs):
        actual_response = asynctest.MagicMock(*args, **kwargs)
        async def resp_mock_effect(request):
            nonlocal actual_response

            # The request's content can be read inside of the handler only. We preserve
            # the data into a conventional field, so that they could be asserted later.
            try:
                request.data = await request.json()
            except json.JSONDecodeError:
                request.data = await request.text()

            # Get a response/error as it was intended (via return_value/side_effect).
            response = actual_response()
            return response

        return asynctest.CoroutineMock(side_effect=resp_mock_effect)
    return resp_maker


@pytest.fixture()
def version_api(resp_mocker, aresponses, hostname, resource):
    result = {'resources': [{
        'name': resource.plural,
        'namespaced': True,
    }]}
    list_mock = resp_mocker(return_value=aiohttp.web.json_response(result))
    aresponses.add(hostname, resource.get_version_url(), 'get', list_mock)


@pytest.fixture()
def stream(fake_vault, resp_mocker, aresponses, hostname, resource, version_api):
    """ A mock for the stream of events as if returned by K8s client. """

    def feed(*args, namespace=None):
        for arg in args:

            # Prepare the stream response pre-rendered (for simplicity, no actual streaming).
            if isinstance(arg, (list, tuple)):
                stream_text = '\n'.join(json.dumps(event) for event in arg)
                stream_resp = aresponses.Response(text=stream_text)
            else:
                stream_resp = arg

            # List is requested for every watch, so we simulate it empty.
            list_data = {'items': [], 'metadata': {'resourceVersion': '0'}}
            list_resp = aiohttp.web.json_response(list_data)
            list_url = resource.get_url(namespace=namespace)

            # The stream is not empty, but is as fed.
            stream_query = {'watch': 'true', 'resourceVersion': '0'}
            stream_url = resource.get_url(namespace=namespace, params=stream_query)

            # Note: `aresponses` excludes a response once it is matched (side-effect-like).
            # So we just accumulate them there, as many as needed.
            aresponses.add(hostname, stream_url, 'get', stream_resp, match_querystring=True)
            aresponses.add(hostname, list_url, 'get', list_resp, match_querystring=True)

    # TODO: One day, find a better way to terminate a ``while-true`` reconnection cycle.
    def close(*, namespace=None):
        """
        A way to stop the stream from reconnecting: say it that the resource version is gone
        (we know a priori that it stops on this condition, and escalates to `infinite_stream`).
        """
        feed([{'type': 'ERROR', 'object': {'code': 410}}], namespace=namespace)

    return Mock(spec_set=['feed', 'close'], feed=feed, close=close)


#
# Mocks for login & checks. Used in specifialised login tests,
# and in all CLI tests (since login is implicit with CLI commands).
#

@pytest.fixture()
def hostname():
    """ A fake hostname to be used in all aiohttp/aresponses tests. """
    return 'fake-host'


@dataclasses.dataclass(frozen=True, eq=False, order=False)
class LoginMocks:
    pykube_in_cluster: Mock = None
    pykube_from_file: Mock = None
    client_in_cluster: Mock = None
    client_from_file: Mock = None


@pytest.fixture()
def login_mocks(mocker):
    """
    Make all client libraries potentially optional, but do not skip the tests:
    skipping the tests is the tests' decision, not this mocking fixture's one.
    """
    kwargs = {}
    try:
        import pykube
    except ImportError:
        pass
    else:
        cfg = pykube.KubeConfig({
            'current-context': 'self',
            'clusters': [{'name': 'self',
                          'cluster': {'server': 'localhost'}}],
            'contexts': [{'name': 'self',
                          'context': {'cluster': 'self', 'namespace': 'default'}}],
        })
        kwargs.update(
            pykube_in_cluster=mocker.patch.object(pykube.KubeConfig, 'from_service_account', return_value=cfg),
            pykube_from_file=mocker.patch.object(pykube.KubeConfig, 'from_file', return_value=cfg),
        )
    try:
        import kubernetes
    except ImportError:
        pass
    else:
        kwargs.update(
            client_in_cluster=mocker.patch.object(kubernetes.config, 'load_incluster_config'),
            client_from_file=mocker.patch.object(kubernetes.config, 'load_kube_config'),
        )
    return LoginMocks(**kwargs)


@pytest.fixture(autouse=True)
def clean_kubernetes_client():
    try:
        import kubernetes
    except ImportError:
        pass  # absent client is already "clean" (or not "dirty" at least).
    else:
        kubernetes.client.configuration.Configuration.set_default(None)


@pytest.fixture()
def fake_vault(mocker, hostname):
    """
    Provide a freshly created and populated authentication vault for every test.

    Most of the tests expect some credentials to be at least provided
    (even if not used). So, we create and set the vault as if every coroutine
    is invoked from the central `operator` method (where it is set normally).

    Any blocking activities are mocked, so that the tests do not hang.
    """
    from kopf.clients import auth

    key = VaultKey('fixture')
    info = ConnectionInfo(server=f'https://{hostname}')
    vault = Vault({key: info})
    token = auth.vault_var.set(vault)
    mocker.patch.object(vault._ready, 'wait_for')
    try:
        yield vault
    finally:
        auth.vault_var.reset(token)

#
# Simulating that Kubernetes client libraries are not installed.
#

def _with_module_present(name: str):
    # If the module is required, it should either be installed,
    # or skip the test: we cannot simulate its presence (unlike its absence).
    yield pytest.importorskip(name)


def _with_module_absent(name: str):

    class ProhibitedImportFinder:
        def find_spec(self, fullname, path, target=None):
            if fullname.split('.')[0] == name:
                raise ImportError("Import is prohibited for tests.")

    try:
        mod_before = importlib.import_module(name)
    except ImportError:
        yield
        return  # nothing to patch & restore.

    # Remove any cached modules.
    preserved = {}
    for fullname, mod in list(sys.modules.items()):
        if fullname.split('.')[0] == name:
            preserved[fullname] = mod
            del sys.modules[fullname]

    # Inject the prohibition for loading this module. And restore when done.
    finder = ProhibitedImportFinder()
    sys.meta_path.insert(0, finder)
    try:
        yield
    finally:
        sys.meta_path.remove(finder)
        sys.modules.update(preserved)

        # Verify if it works and that we didn't break the importing machinery.
        mod_after = importlib.import_module(name)
        assert mod_after is mod_before


@pytest.fixture(params=[True], ids=['with-client'])  # for hinting suffixes
def kubernetes():
    yield from _with_module_present('kubernetes')


@pytest.fixture(params=[False], ids=['no-client'])  # for hinting suffixes
def no_kubernetes():
    yield from _with_module_absent('kubernetes')


@pytest.fixture(params=[False, True], ids=['no-client', 'with-client'])
def any_kubernetes(request):
    if request.param:
        yield from _with_module_present('kubernetes')
    else:
        yield from _with_module_absent('kubernetes')


@pytest.fixture(params=[True], ids=['with-pykube'])  # for hinting suffixes
def pykube():
    yield from _with_module_present('pykube')


@pytest.fixture(params=[False], ids=['no-pykube'])  # for hinting suffixes
def no_pykube():
    yield from _with_module_absent('pykube')


@pytest.fixture(params=[False, True], ids=['no-pykube', 'with-pykube'])
def any_pykube(request):
    if request.param:
        yield from _with_module_present('pykube')
    else:
        yield from _with_module_absent('pykube')

#
# Helpers for the timing checks.
#

@pytest.fixture()
def timer():
    return Timer()


class Timer(object):
    """
    A helper context manager to measure the time of the code-blocks.
    Also, supports direct comparison with time-deltas and the numbers of seconds.

    Usage:

        with Timer() as timer:
            do_something()
            print(f"Executing for {timer.seconds}s already.")
            do_something_else()

        print(f"Executed in {timer.seconds}s.")
        assert timer < 5.0
    """

    def __init__(self):
        super().__init__()
        self._ts = None
        self._te = None

    @property
    def seconds(self):
        if self._ts is None:
            return None
        elif self._te is None:
            return time.perf_counter() - self._ts
        else:
            return self._te - self._ts

    def __repr__(self):
        status = 'new' if self._ts is None else 'running' if self._te is None else 'finished'
        return f'<Timer: {self.seconds}s ({status})>'

    def __enter__(self):
        self._ts = time.perf_counter()
        self._te = None
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._te = time.perf_counter()

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return self.__exit__(exc_type, exc_val, exc_tb)

    def __int__(self):
        return int(self.seconds)

    def __float__(self):
        return float(self.seconds)

#
# Helpers for the logging checks.
#


@pytest.fixture()
def logstream(caplog):
    """ Prefixing is done at the final output. We have to intercept it. """

    logger = logging.getLogger()
    handlers = list(logger.handlers)

    # Setup all log levels of sub-libraries. A sife-effect: the handlers are also added.
    configure(verbose=True)

    # Remove any stream handlers added in the step above. But keep the caplog's handlers.
    for handler in list(logger.handlers):
        if isinstance(handler, logging.StreamHandler) and handler.stream is sys.stderr:
            logger.removeHandler(handler)

    # Inject our stream-intercepting handler.
    stream = io.StringIO()
    handler = logging.StreamHandler(stream)
    formatter = ObjectPrefixingTextFormatter('prefix %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    try:
        with caplog.at_level(logging.DEBUG):
            yield stream
    finally:
        logger.removeHandler(handler)
        logger.handlers[:] = handlers  # undo `configure()`


@pytest.fixture()
def assert_logs(caplog):
    """
    A function to assert the logs are present (by pattern).

    The listed message patterns MUST be present, in the order specified.
    Some other log messages can also be present, but they are ignored.
    """
    def assert_logs_fn(patterns, prohibited=[], strict=False):
        __traceback_hide__ = True
        remaining_patterns = list(patterns)
        for message in caplog.messages:
            # The expected pattern is at position 0.
            # Looking-ahead: if one of the following patterns matches, while the
            # 0th does not, then the log message is missing, and we fail the test.
            for idx, pattern in enumerate(remaining_patterns):
                m = re.search(pattern, message)
                if m:
                    if idx == 0:
                        remaining_patterns[:1] = []
                        break  # out of `remaining_patterns` cycle
                    else:
                        skipped_patterns = remaining_patterns[:idx]
                        raise AssertionError(f"Few patterns were skipped: {skipped_patterns!r}")
                elif strict:
                    raise AssertionError(f"Unexpected log message: {message!r}")

            # Check that the prohibited patterns do not appear in any message.
            for pattern in prohibited:
                m = re.search(pattern, message)
                if m:
                    raise AssertionError(f"Prohibited log pattern found: {message!r} ~ {pattern!r}")

        # If all patterns have been matched in order, we are done.
        # if some are left, but the messages are over, then we fail.
        if remaining_patterns:
            raise AssertionError(f"Few patterns were missed: {remaining_patterns!r}")

    return assert_logs_fn


#
# Helpers for asyncio checks.
#
@pytest.fixture(autouse=True)
def _no_asyncio_pending_tasks(event_loop):
    """
    Ensure there are no unattended asyncio tasks after the test.

    It looks  both in the test's main event-loop, and in all other event-loops,
    such as the background thread of `KopfRunner` (used in e2e tests).

    Current solution uses some internals of asyncio, since there is no public
    interface for that. The warnings are printed only at the end of pytest.

    An alternative way: set event-loop's exception handler, force garbage
    collection after every test, and check messages from `asyncio.Task.__del__`.
    This, however, requires intercepting all event-loop creation in the code.
    """

    # See `asyncio.all_tasks()` implementation for reference.
    before = {t for t in list(asyncio.tasks._all_tasks) if not t.done()}

    # Run the test.
    yield

    # Let the pytest-asyncio's async2sync wrapper to finish all callbacks. Otherwise, it raises:
    #   <Task pending name='Task-2' coro=<<async_generator_athrow without __name__>()>>
    event_loop.run_until_complete(asyncio.sleep(0))

    # Detect all leftover tasks.
    after = {t for t in list(asyncio.tasks._all_tasks) if not t.done()}
    remains = after - before
    if remains:
        pytest.fail(f"Unattended asyncio tasks detected: {remains!r}")



import asyncio

import aiohttp.web
import async_timeout
import pytest

import kopf
from kopf.reactor.observation import process_discovered_resource_event
from kopf.structs.bodies import RawBody, RawEvent
from kopf.structs.references import NAMESPACES, Insights, Resource

# Implementation awareness: the events only trigger the re-scan, so the fields can be reduced
# to only the group name which is being rescanned. Other fields are ignored in the events.
# The actual data is taken from the API. It is tested elsewhere, so we rely on its correctness here.
GROUP1_EVENT = RawBody(spec={'group': 'group1'})


@pytest.fixture()
def core_mock(resp_mocker, aresponses, hostname):
    mock = resp_mocker(return_value=aiohttp.web.json_response({'versions': ['v1']}))
    aresponses.add(hostname, '/api', 'get', mock)
    return mock


@pytest.fixture()
def apis_mock(resp_mocker, aresponses, hostname):
    mock = resp_mocker(return_value=aiohttp.web.json_response({'groups': [
        {
            'name': 'group1',
            'preferredVersion': {'version': 'version1'},
            'versions': [{'version': 'version1'}],
        },
    ]}))
    aresponses.add(hostname, '/apis', 'get', mock)
    return mock


@pytest.fixture()
def corev1_mock(resp_mocker, aresponses, hostname, core_mock):
    mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': [
        {
            'kind': 'Namespace',
            'name': 'namespaces',
            'singularName': 'namespace',
            'namespaced': False,
            'categories': [],
            'shortNames': [],
            'verbs': ['list', 'watch', 'patch'],
        },
    ]}))
    aresponses.add(hostname, '/api/v1', 'get', mock)
    return mock


@pytest.fixture()
def group1_mock(resp_mocker, aresponses, hostname, apis_mock):
    mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': [
        {
            'kind': 'kind1',
            'name': 'plural1',
            'singularName': 'singular1',
            'namespaced': True,
            'categories': ['category1a', 'category1b'],
            'shortNames': ['shortname1a', 'shortname1b'],
            'verbs': ['list', 'watch', 'patch'],
        },
    ]}))
    aresponses.add(hostname, '/apis/group1/version1', 'get', mock)
    return mock


@pytest.fixture()
def group1_empty_mock(resp_mocker, aresponses, hostname, apis_mock):
    mock = resp_mocker(return_value=aiohttp.web.json_response({'resources': []}))
    aresponses.add(hostname, '/apis/group1/version1', 'get', mock)
    return mock


@pytest.fixture()
def group1_404mock(resp_mocker, aresponses, hostname, apis_mock):
    mock = resp_mocker(return_value=aresponses.Response(status=404))
    aresponses.add(hostname, '/apis/group1/version1', 'get', mock)
    return mock


@pytest.fixture(params=[
    kopf.on.event, kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def handlers(request, registry):
    @request.param('group1', 'version1', 'plural1')
    def fn(**_): ...



async def test_initial_listing_is_ignored(registry, apis_mock, group1_mock):
    insights = Insights()
    e1 = RawEvent(type=None, object=RawBody(spec={'group': 'group1'}))

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, replenished=asyncio.Event())

    task = asyncio.create_task(delayed_injection(0))
    with pytest.raises(asyncio.TimeoutError):
        async with async_timeout.timeout(0.1) as timeout:
            async with insights.revised:
                await insights.revised.wait()
    await task
    assert timeout.expired
    assert not insights.resources
    assert not apis_mock.called
    assert not group1_mock.called


@pytest.mark.usefixtures('handlers')
@pytest.mark.parametrize('etype', ['ADDED', 'MODIFIED'])
async def test_followups_for_addition(registry, apis_mock, group1_mock, timer, etype):
    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))
    r1 = Resource(group='group1', version='version1', plural='plural1')
    insights = Insights()

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, replenished=asyncio.Event())

    task = asyncio.create_task(delayed_injection(0.1))
    async with timer, async_timeout.timeout(1.0):
        async with insights.revised:
            await insights.revised.wait()
    await task
    assert 0.1 < timer.seconds < 1.0
    assert insights.resources == {r1}
    assert apis_mock.called
    assert group1_mock.called


@pytest.mark.usefixtures('handlers')
@pytest.mark.parametrize('etype', ['ADDED', 'MODIFIED', 'DELETED'])
async def test_followups_for_deletion_of_resource(registry, apis_mock, group1_empty_mock, timer, etype):
    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))
    r1 = Resource(group='group1', version='version1', plural='plural1')
    insights = Insights()
    insights.resources.add(r1)

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, replenished=asyncio.Event())

    task = asyncio.create_task(delayed_injection(0.1))
    async with timer, async_timeout.timeout(1.0):
        async with insights.revised:
            await insights.revised.wait()
    await task
    assert 0.1 < timer.seconds < 1.0
    assert not insights.resources
    assert apis_mock.called
    assert group1_empty_mock.called


@pytest.mark.usefixtures('handlers')
@pytest.mark.parametrize('etype', ['ADDED', 'MODIFIED', 'DELETED'])
async def test_followups_for_deletion_of_group(registry, apis_mock, group1_404mock, timer, etype):
    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))
    r1 = Resource(group='group1', version='version1', plural='plural1')
    insights = Insights()
    insights.resources.add(r1)

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, replenished=asyncio.Event())

    task = asyncio.create_task(delayed_injection(0.1))
    async with timer, async_timeout.timeout(1.0):
        async with insights.revised:
            await insights.revised.wait()
    await task
    assert 0.1 < timer.seconds < 1.0
    assert not insights.resources
    assert apis_mock.called
    assert group1_404mock.called


@pytest.mark.usefixtures('handlers')
@pytest.mark.parametrize('etype', ['DELETED'])
async def test_followups_for_deletion_of_group(registry, apis_mock, group1_404mock, timer, etype):
    e1 = RawEvent(type=etype, object=RawBody(spec={'group': 'group1'}))
    r1 = Resource(group='group1', version='version1', plural='plural1')
    insights = Insights()
    insights.resources.add(r1)

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, replenished=asyncio.Event())

    task = asyncio.create_task(delayed_injection(0.1))
    async with timer, async_timeout.timeout(1.0):
        async with insights.revised:
            await insights.revised.wait()
    await task
    assert 0.1 < timer.seconds < 1.0
    assert not insights.resources
    assert apis_mock.called
    assert group1_404mock.called


@pytest.mark.parametrize('etype', ['DELETED'])
async def test_backbone_is_filled(registry, core_mock, corev1_mock, timer, etype):
    e1 = RawEvent(type=etype, object=RawBody(spec={'group': ''}))
    insights = Insights()

    async def delayed_injection(delay: float):
        await asyncio.sleep(delay)
        await process_discovered_resource_event(
            insights=insights, raw_event=e1, registry=registry, replenished=asyncio.Event())

    task = asyncio.create_task(delayed_injection(0.1))
    async with timer, async_timeout.timeout(1.0):
        await insights.backbone.wait_for(NAMESPACES)
    await task
    assert 0.1 < timer.seconds < 1.0
    assert NAMESPACES in insights.backbone
    assert core_mock.called
    assert corev1_mock.called



"""
Conversion of low-level events to high-level causes, and handling them.

These functions are invoked from the queueing module `kopf.reactor.queueing`,
which are the actual event loop of the operator process.

The conversion of the low-level events to the high-level causes is done by
checking the object's state and comparing it to the preserved last-seen state.

The framework itself makes the necessary changes to the object, -- such as the
finalizers attachment, last-seen state updates, and handler status tracking, --
thus provoking the low-level watch-events and additional queueing calls.
But these internal changes are filtered out from the cause detection
and therefore do not trigger the user-defined handlers.
"""
import asyncio
import time
from typing import Collection, Optional, Tuple

from kopf.engines import loggers, posting
from kopf.reactor import causation, daemons, effects, handling, lifecycles, registries
from kopf.storage import finalizers, states
from kopf.structs import bodies, configuration, containers, diffs, \
                         handlers as handlers_, patches, references


async def process_resource_event(
        lifecycle: lifecycles.LifeCycleFn,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        memories: containers.ResourceMemories,
        resource: references.Resource,
        raw_event: bodies.RawEvent,
        replenished: asyncio.Event,
        event_queue: posting.K8sEventQueue,
) -> None:
    """
    Handle a single custom object low-level watch-event.

    Convert the low-level events, as provided by the watching/queueing tasks,
    to the high-level causes, and then call the cause-handling logic.
    """

    # Recall what is stored about that object. Share it in little portions with the consumers.
    # And immediately forget it if the object is deleted from the cluster (but keep in memory).
    raw_type, raw_body = raw_event['type'], raw_event['object']
    memory = await memories.recall(raw_body, noticed_by_listing=raw_type is None)
    if memory.live_fresh_body is not None:
        memory.live_fresh_body._replace_with(raw_body)
    if raw_type == 'DELETED':
        await memories.forget(raw_body)

    # Convert to a heavy mapping-view wrapper only now, when heavy processing begins.
    # Raw-event streaming, queueing, and batching use regular lightweight dicts.
    # Why here? 1. Before it splits into multiple causes & handlers for the same object's body;
    # 2. After it is batched (queueing); 3. While the "raw" parsed JSON is still known;
    # 4. Same as where a patch object of a similar wrapping semantics is created.
    body = memory.live_fresh_body if memory.live_fresh_body is not None else bodies.Body(raw_body)
    patch = patches.Patch()

    # Throttle the non-handler-related errors. The regular event watching/batching continues
    # to prevent queue overfilling, but the processing is skipped (events are ignored).
    # Choice of place: late enough to have a per-resource memory for a throttler; also, a logger.
    # But early enough to catch environment errors from K8s API, and from most of the complex code.
    async with effects.throttled(
        throttler=memory.error_throttler,
        logger=loggers.LocalObjectLogger(body=body, settings=settings),
        delays=settings.batching.error_delays,
        wakeup=replenished,
    ) as should_run:
        if should_run:

            # Each object has its own prefixed logger, to distinguish parallel handling.
            logger = loggers.ObjectLogger(body=body, settings=settings)
            posting.event_queue_loop_var.set(asyncio.get_running_loop())
            posting.event_queue_var.set(event_queue)  # till the end of this object's task.

            # Do the magic -- do the job.
            delays, matched = await process_resource_causes(
                lifecycle=lifecycle,
                registry=registry,
                settings=settings,
                resource=resource,
                raw_event=raw_event,
                body=body,
                patch=patch,
                logger=logger,
                memory=memory,
            )

            # Whatever was done, apply the accumulated changes to the object, or sleep-n-touch for delays.
            # But only once, to reduce the number of API calls and the generated irrelevant events.
            # And only if the object is at least supposed to exist (not "GONE"), even if actually does not.
            if raw_event['type'] != 'DELETED':
                applied = await effects.apply(
                    settings=settings,
                    resource=resource,
                    body=body,
                    patch=patch,
                    logger=logger,
                    delays=delays,
                    replenished=replenished,
                )
                if applied and matched:
                    logger.debug(f"Handling cycle is finished, waiting for new changes since now.")


async def process_resource_causes(
        lifecycle: lifecycles.LifeCycleFn,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        resource: references.Resource,
        raw_event: bodies.RawEvent,
        body: bodies.Body,
        patch: patches.Patch,
        logger: loggers.ObjectLogger,
        memory: containers.ResourceMemory,
) -> Tuple[Collection[float], bool]:

    finalizer = settings.persistence.finalizer
    extra_fields = (
        registry._resource_watching.get_extra_fields(resource=resource) |
        registry._resource_changing.get_extra_fields(resource=resource) |
        registry._resource_spawning.get_extra_fields(resource=resource))
    old = settings.persistence.diffbase_storage.fetch(body=body)
    new = settings.persistence.diffbase_storage.build(body=body, extra_fields=extra_fields)
    old = settings.persistence.progress_storage.clear(essence=old) if old is not None else None
    new = settings.persistence.progress_storage.clear(essence=new) if new is not None else None
    diff = diffs.diff(old, new)

    # Detect what are we going to do on this processing cycle.
    resource_watching_cause = causation.detect_resource_watching_cause(
        raw_event=raw_event,
        resource=resource,
        logger=logger,
        patch=patch,
        body=body,
        memo=memory.memo,
    ) if registry._resource_watching.has_handlers(resource=resource) else None

    resource_spawning_cause = causation.detect_resource_spawning_cause(
        resource=resource,
        logger=logger,
        patch=patch,
        body=body,
        memo=memory.memo,
        reset=bool(diff),  # only essential changes reset idling, not every event
    ) if registry._resource_spawning.has_handlers(resource=resource) else None

    resource_changing_cause = causation.detect_resource_changing_cause(
        finalizer=finalizer,
        raw_event=raw_event,
        resource=resource,
        logger=logger,
        patch=patch,
        body=body,
        old=old,
        new=new,
        diff=diff,
        memo=memory.memo,
        initial=memory.noticed_by_listing and not memory.fully_handled_once,
    ) if registry._resource_changing.has_handlers(resource=resource) else None

    # If there are any handlers for this resource kind in general, but not for this specific object
    # due to filters, then be blind to it, store no state, and log nothing about the handling cycle.
    if (resource_changing_cause is not None and
        not registry._resource_changing.prematch(cause=resource_changing_cause)):
        resource_changing_cause = None

    # Block the object from deletion if we have anything to do in its end of life:
    # specifically, if there are daemons to kill or mandatory on-deletion handlers to call.
    # The high-level handlers are prevented if this event cycle is dedicated to the finalizer.
    # The low-level handlers (on-event spying & daemon spawning) are still executed asap.
    deletion_is_ongoing = finalizers.is_deletion_ongoing(body=body)
    deletion_is_blocked = finalizers.is_deletion_blocked(body=body, finalizer=finalizer)
    deletion_must_be_blocked = (
        (resource_spawning_cause is not None and
         registry._resource_spawning.requires_finalizer(
             cause=resource_spawning_cause,
             excluded=memory.forever_stopped,
         ))
        or
        (resource_changing_cause is not None and
         registry._resource_changing.requires_finalizer(
             cause=resource_changing_cause,
         )))

    if deletion_must_be_blocked and not deletion_is_blocked and not deletion_is_ongoing:
        logger.debug("Adding the finalizer, thus preventing the actual deletion.")
        finalizers.block_deletion(body=body, patch=patch, finalizer=finalizer)
        resource_changing_cause = None  # prevent further high-level processing this time

    if not deletion_must_be_blocked and deletion_is_blocked:
        logger.debug("Removing the finalizer, as there are no handlers requiring it.")
        finalizers.allow_deletion(body=body, patch=patch, finalizer=finalizer)
        resource_changing_cause = None  # prevent further high-level processing this time

    # Invoke all the handlers that should or could be invoked at this processing cycle.
    # The low-level spies go ASAP always. However, the daemons are spawned before the high-level
    # handlers and killed after them: the daemons should live throughout the full object lifecycle.
    if resource_watching_cause is not None:
        await process_resource_watching_cause(
            lifecycle=lifecycles.all_at_once,
            registry=registry,
            settings=settings,
            cause=resource_watching_cause,
        )

    resource_spawning_delays: Collection[float] = []
    if resource_spawning_cause is not None:
        resource_spawning_delays = await process_resource_spawning_cause(
            registry=registry,
            settings=settings,
            memory=memory,
            cause=resource_spawning_cause,
        )

    resource_changing_delays: Collection[float] = []
    if resource_changing_cause is not None:
        resource_changing_delays = await process_resource_changing_cause(
            lifecycle=lifecycle,
            registry=registry,
            settings=settings,
            memory=memory,
            cause=resource_changing_cause,
        )

    # Release the object if everything is done, and it is marked for deletion.
    # But not when it has already gone.
    if deletion_is_ongoing and deletion_is_blocked \
            and not resource_spawning_delays \
            and not resource_changing_delays:
        logger.debug("Removing the finalizer, thus allowing the actual deletion.")
        finalizers.allow_deletion(body=body, patch=patch, finalizer=finalizer)

    delays = list(resource_spawning_delays) + list(resource_changing_delays)
    return (delays, resource_changing_cause is not None)


async def process_resource_watching_cause(
        lifecycle: lifecycles.LifeCycleFn,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        cause: causation.ResourceWatchingCause,
) -> None:
    """
    Handle a received event, log but ignore all errors.

    This is a lightweight version of the cause handling, but for the raw events,
    without any progress persistence. Multi-step calls are also not supported.
    If the handler fails, it fails and is never retried.

    Note: K8s-event posting is skipped for `kopf.on.event` handlers,
    as they should be silent. Still, the messages are logged normally.
    """
    handlers = registry._resource_watching.get_handlers(cause=cause)
    outcomes = await handling.execute_handlers_once(
        lifecycle=lifecycle,
        settings=settings,
        handlers=handlers,
        cause=cause,
        state=states.State.from_scratch().with_handlers(handlers),
        default_errors=handlers_.ErrorsMode.IGNORED,
    )

    # Store the results, but not the handlers' progress.
    states.deliver_results(outcomes=outcomes, patch=cause.patch)


async def process_resource_spawning_cause(
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        memory: containers.ResourceMemory,
        cause: causation.ResourceSpawningCause,
) -> Collection[float]:
    """
    Spawn/kill all the background tasks of a resource.

    The spawning and killing happens in parallel with the resource-changing
    handlers invocation (even if it takes few cycles). For this, the signal
    to terminate is sent to the daemons immediately, but the actual check
    of their shutdown is performed only when all the on-deletion handlers
    have succeeded (or after they were invoked if they are optional;
    or immediately if there were no on-deletion handlers to invoke at all).

    The resource remains blocked by the finalizers until all the daemons exit
    (except those marked as tolerating being orphaned).
    """

    # Refresh the up-to-date body & essential timestamp for all the daemons/timers.
    if memory.live_fresh_body is None:
        memory.live_fresh_body = cause.body
    if cause.reset:
        memory.idle_reset_time = time.monotonic()

    if finalizers.is_deletion_ongoing(cause.body):
        stopping_delays = await daemons.stop_resource_daemons(
            settings=settings,
            daemons=memory.running_daemons,
        )
        return stopping_delays

    else:
        handlers = registry._resource_spawning.get_handlers(
            cause=cause,
            excluded=memory.forever_stopped,
        )
        spawning_delays = await daemons.spawn_resource_daemons(
            settings=settings,
            daemons=memory.running_daemons,
            cause=cause,
            memory=memory,
            handlers=handlers,
        )
        matching_delays = await daemons.match_resource_daemons(
            settings=settings,
            daemons=memory.running_daemons,
            handlers=handlers,
        )
        return list(spawning_delays) + list(matching_delays)


async def process_resource_changing_cause(
        lifecycle: lifecycles.LifeCycleFn,
        registry: registries.OperatorRegistry,
        settings: configuration.OperatorSettings,
        memory: containers.ResourceMemory,
        cause: causation.ResourceChangingCause,
) -> Collection[float]:
    """
    Handle a detected cause, as part of the bigger handler routine.
    """
    logger = cause.logger
    patch = cause.patch  # TODO get rid of this alias
    body = cause.body  # TODO get rid of this alias
    delays: Collection[float] = []
    done: Optional[bool] = None
    skip: Optional[bool] = None

    # Regular causes invoke the handlers.
    if cause.reason in handlers_.HANDLER_REASONS:
        title = handlers_.TITLES.get(cause.reason, repr(cause.reason))

        resource_registry = registry._resource_changing
        owned_handlers = resource_registry.get_resource_handlers(resource=cause.resource)
        cause_handlers = resource_registry.get_handlers(cause=cause)
        storage = settings.persistence.progress_storage
        state = states.State.from_storage(body=cause.body, storage=storage, handlers=owned_handlers)
        state = state.with_purpose(cause.reason).with_handlers(cause_handlers)

        # Report the causes that have been superseded (intercepted, overridden) by the current one.
        # The mix-in causes (i.e. resuming) is re-purposed if its handlers are still selected.
        # To the next cycle, all extras are purged or re-purposed, so the message does not repeat.
        for extra_reason, counters in state.extras.items():  # usually 0..1 items, rarely 2+.
            extra_title = handlers_.TITLES.get(extra_reason, repr(extra_reason))
            logger.info(f"{extra_title.capitalize()} is superseded by {title.lower()}: "
                        f"{counters.success} succeeded; "
                        f"{counters.failure} failed; "
                        f"{counters.running} left to the moment.")
            state = state.with_purpose(purpose=cause.reason, handlers=cause_handlers)

        # Purge the now-irrelevant handlers if they were not re-purposed (extras are recalculated!).
        # The current cause continues afterwards, and overrides its own pre-purged handler states.
        # TODO: purge only the handlers that fell out of current purpose; but it is not critical
        if state.extras:
            state.purge(body=cause.body, patch=cause.patch,
                        storage=storage, handlers=owned_handlers)

        # Inform on the current cause/event on every processing cycle. Even if there are
        # no handlers -- to show what has happened and why the diff-base is patched.
        logger.debug(f"{title.capitalize()} is in progress: %r", body)
        if cause.diff and cause.old is not None and cause.new is not None:
            logger.debug(f"{title.capitalize()} diff: %r", cause.diff)

        if cause_handlers:
            outcomes = await handling.execute_handlers_once(
                lifecycle=lifecycle,
                settings=settings,
                handlers=cause_handlers,
                cause=cause,
                state=state,
            )
            state = state.with_outcomes(outcomes)
            state.store(body=cause.body, patch=cause.patch, storage=storage)
            states.deliver_results(outcomes=outcomes, patch=cause.patch)

            if state.done:
                counters = state.counts  # calculate only once
                logger.info(f"{title.capitalize()} is processed: "
                            f"{counters.success} succeeded; "
                            f"{counters.failure} failed.")
                state.purge(body=cause.body, patch=cause.patch,
                            storage=storage, handlers=owned_handlers)

            done = state.done
            delays = state.delays
        else:
            skip = True

    # Regular causes also do some implicit post-handling when all handlers are done.
    if done or skip:
        if cause.new is not None and cause.old != cause.new:
            settings.persistence.diffbase_storage.store(body=body, patch=patch, essence=cause.new)

        # Once all handlers have succeeded at least once for any reason, or if there were none,
        # prevent further resume-handlers (which otherwise happens on each watch-stream re-listing).
        memory.fully_handled_once = True

    # Informational causes just print the log lines.
    if cause.reason == handlers_.Reason.GONE:
        logger.debug("Deleted, really deleted, and we are notified.")

    if cause.reason == handlers_.Reason.FREE:
        logger.debug("Deletion, but we are done with it, and we do not care.")

    if cause.reason == handlers_.Reason.NOOP:
        logger.debug("Something has changed, but we are not interested (the essence is the same).")

    # The delay is then consumed by the main handling routine (in different ways).
    return delays



import pytest

from kopf.engines.peering import keepalive


class StopInfiniteCycleException(Exception):
    pass


async def test_background_task_runs(mocker, settings, namespaced_peering_resource):
    touch_mock = mocker.patch('kopf.engines.peering.touch')

    sleep_mock = mocker.patch('asyncio.sleep')
    sleep_mock.side_effect = [None, None, StopInfiniteCycleException]

    randint_mock = mocker.patch('random.randint')
    randint_mock.side_effect = [7, 5, 9]

    settings.peering.lifetime = 33
    with pytest.raises(StopInfiniteCycleException):
        await keepalive(settings=settings, identity='id',
                        resource=namespaced_peering_resource, namespace='namespace')

    assert randint_mock.call_count == 3  # only to be sure that we test the right thing
    assert sleep_mock.call_count == 3
    assert sleep_mock.call_args_list[0][0][0] == 33 - 7
    assert sleep_mock.call_args_list[1][0][0] == 33 - 5
    assert sleep_mock.call_args_list[2][0][0] == 33 - 9

    assert touch_mock.call_count == 4  # 3 updates + 1 clean-up



"""
K8s API errors.

The underlying client library (now, ``aiohttp``) can be replaced in the future.
We cannot rely on embedding its exceptions all over the code in the framework.
Hence, we have our own hierarchy of exceptions for K8s API errors.

Low-level errors, such as the network connectivity issues, SSL/HTTPS issues,
etc, are escalated from the client library as is, since they are related not
to the domain of K8s API, but rather to the networking and encryption.

The original errors of the client library are chained as the causes of our own
specialised errors -- for better explainability of errors in the stack traces.

Some selected reasons of K8s API errors are made into their own classes,
so that they could be intercepted and handled in other places of the framework.
All other reasons are raised as the base error class and are indistinguishable
from each other (except via the exception's fields).

Unlike the underlying client library's errors, the K8s API errors contain more
information about the reasons -- as provided by K8s API in its response bodies,
not guessed only by HTTP statuses alone.

These errors are not exposed to the users, and the users cannot catch them
with ``except:`` clauses. The users can only see these errors in the logs
as the reasons of failures. However, the errors are exposed to other packages.
"""
import collections.abc
import json
from typing import Any, Collection, Optional

import aiohttp
from typing_extensions import Literal, TypedDict


class RawStatusCause(TypedDict):
    field: str
    reason: str
    message: str


class RawStatusDetails(TypedDict):
    name: str
    uid: str
    retryAfterSeconds: int
    kind: str
    group: str
    causes: Collection[RawStatusCause]


# https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#status-v1-meta
class RawStatus(TypedDict):
    apiVersion: str
    kind: Literal["Status"]
    code: int
    status: Literal["Success", "Failure"]
    reason: str
    message: str
    details: RawStatusDetails


class APIError(Exception):

    def __init__(
            self,
            payload: Optional[RawStatus],
            *,
            status: int,
    ) -> None:
        message = payload.get('message') if payload else None
        super().__init__(message, payload)
        self._status = status
        self._payload = payload

    @property
    def status(self) -> int:
        return self._status

    @property
    def code(self) -> Optional[int]:
        return self._payload.get('code') if self._payload else None

    @property
    def message(self) -> Optional[str]:
        return self._payload.get('message') if self._payload else None

    @property
    def details(self) -> Optional[RawStatusDetails]:
        return self._payload.get('details') if self._payload else None


class APIUnauthorizedError(APIError):
    pass


class APIForbiddenError(APIError):
    pass


class APINotFoundError(APIError):
    pass


async def check_response(
        response: aiohttp.ClientResponse,
) -> None:
    """
    Check for specialised K8s errors, and raise with extended information.
    """
    if response.status >= 400:

        # Read the response's body before it is closed by raise_for_status().
        payload: Optional[RawStatus]
        try:
            payload = await response.json()
        except (json.JSONDecodeError, aiohttp.ContentTypeError, aiohttp.ClientConnectionError):
            payload = None

        # Better be safe: who knows which sensitive information can be dumped unless kind==Status.
        if not isinstance(payload, collections.abc.Mapping) or payload.get('kind') != 'Status':
            payload = None

        cls = (
            APIUnauthorizedError if response.status == 401 else
            APIForbiddenError if response.status == 403 else
            APINotFoundError if response.status == 404 else
            APIError
        )

        # Raise the framework-specific error while keeping the original error in scope.
        # This call also closes the response's body, so it cannot be read afterwards.
        try:
            response.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise cls(payload, status=response.status) from e


async def parse_response(
        response: aiohttp.ClientResponse,
) -> Any:
    """
    Check the response for errors, and either raise or returned the parsed data.
    """
    await check_response(response)
    payload = await response.json()
    return payload



"""
Helpers for orchestrating asyncio tasks.

These utilities only support tasks, not more generic futures, coroutines,
or other awaitables. In most case where we use it, we need specifically tasks,
as we not only wait for them, but also cancel them.

Anyway, ``asyncio`` wraps all awaitables and coroutines into tasks on almost
all function calls with multiple awaiables (e.g. :func:`asyncio.wait`),
so there is no added overhead; intstead, the implicit overhead is made explicit.
"""
import asyncio
import logging
import sys
from typing import TYPE_CHECKING, Any, Awaitable, Collection, Coroutine, \
                   Generator, Optional, Set, Tuple, TypeVar, Union, cast

_T = TypeVar('_T')

# A workaround for a difference in tasks at runtime and type-checking time.
# Otherwise, at runtime: TypeError: 'type' object is not subscriptable.
if TYPE_CHECKING:
    Future = asyncio.Future[Any]
    Task = asyncio.Task[Any]
else:
    Future = asyncio.Future
    Task = asyncio.Task

# Accept `name=` always, but simulate it for Python 3.7 to do nothing.
if sys.version_info >= (3, 8):
    create_task = asyncio.create_task
else:
    def create_task(
            coro: Union[Generator[Any, None, _T], Awaitable[_T]],
            *,
            name: Optional[str] = None,  # noqa
    ) -> Task:
        return asyncio.create_task(coro)


async def guard(
        coro: Coroutine[Any, Any, Any],
        name: str,
        *,
        flag: Optional[asyncio.Event] = None,
        finishable: bool = False,
        cancellable: bool = False,
        logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None,
) -> None:
    """
    A guard for a presumably eternal (never-finishing) task.

    An "eternal" task is a task that never exits unless explicitly cancelled.
    If it does, this is a misbehaviour that is logged. Errors are always logged.
    Cancellations are also logged except if the task is said to be cancellable.

    It is used for background tasks that are started but never awaited/checked,
    so that the errors are not escalated properly; or if they are occasionally
    awaited/checked with a significant delay after an error possibly happend,
    but needs to be logged as soon as it happens.
    """
    capname = name.capitalize()

    # Guarded tasks can have prerequisites, which are set in other tasks.
    if flag is not None:
        await flag.wait()

    try:
        await coro
    except asyncio.CancelledError:
        if logger is not None and not cancellable:
            logger.debug(f"{capname} is cancelled.")
        raise
    except Exception as e:
        if logger is not None:
            logger.exception(f"{capname} has failed: %s", e)
        raise
    else:
        if logger is not None and not finishable:
            logger.warning(f"{capname} has finished unexpectedly.")


def create_guarded_task(
        coro: Coroutine[Any, Any, Any],
        name: str,
        *,
        flag: Optional[asyncio.Event] = None,
        finishable: bool = False,
        cancellable: bool = False,
        logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None,
) -> Task:
    """
    Create a guarded eternal task. See :func:`guard` for explanation.

    This is only a shortcut for named task creation (name is used in 2 places).
    """
    return create_task(
        name=name,
        coro=guard(
            name=name,
            coro=coro,
            flag=flag,
            finishable=finishable,
            cancellable=cancellable,
            logger=logger))


async def wait(
        tasks: Collection[Task],
        *,
        timeout: Optional[float] = None,
        return_when: Any = asyncio.ALL_COMPLETED,
) -> Tuple[Set[Task], Set[Task]]:
    """
    A safer version of :func:`asyncio.wait` -- does not fail on an empty list.
    """
    if not tasks:
        return set(), set()
    done, pending = await asyncio.wait(tasks, timeout=timeout, return_when=return_when)
    return cast(Set[Task], done), cast(Set[Task], pending)


async def stop(
        tasks: Collection[Task],
        *,
        title: str,
        quiet: bool = False,
        cancelled: bool = False,
        interval: Optional[float] = None,
        logger: Optional[Union[logging.Logger, logging.LoggerAdapter]] = None,
) -> Tuple[Set[Task], Set[Task]]:
    """
    Cancel the tasks and wait for them to finish; log if some are stuck.

    By default, all stopping activities are logged. In the quiet mode, logs
    only the tasks that are stuck longer than the designated polling interval,
    and their eventual exit if they were reported as stuck at least once.

    If the interval is not set, no polling is performed, and the stopping
    should happen in one iteration (even if it is going to take an eternity).

    The stopping itself does not have timeouts. It always ends either with
    the tasks stopped/exited, or with the stop-routine itself being cancelled.

    For better logging only, if the stopping routine is marked as performing
    the cancellation already (via ``cancelled=True``), then the cancellation
    of the stopping routine is considered as "double-cancelling".
    This does not affect the behaviour, but only the log messages.
    """
    captitle = title.capitalize()

    if not tasks:
        if logger is not None and not quiet:
            logger.debug(f"{captitle} tasks stopping is skipped: no tasks given.")
        return set(), set()

    for task in tasks:
        task.cancel()

    iterations = 0
    done_ever: Set[Task] = set()
    pending: Set[Task] = set(tasks)
    while pending:
        iterations += 1

        # If the waiting (current) task is cancelled before the wait is over,
        # propagate the cancellation to all the awaited (sub-) tasks, and let them finish.
        try:
            done_now, pending = await wait(pending, timeout=interval)
        except asyncio.CancelledError:
            # If the waiting (current) task is cancelled while propagating the cancellation
            # (i.e. double-cancelled), let it fail without graceful cleanup. It is urgent, it seems.
            pending = {task for task in tasks if not task.done()}
            if logger is not None and (not quiet or pending or iterations > 1):
                are = 'are' if not pending else 'are not'
                why = 'double-cancelling at stopping' if cancelled else 'cancelling at stopping'
                logger.debug(f"{captitle} tasks {are} stopped: {why}; tasks left: {pending!r}")
            raise  # the repeated cancellation, handled specially.
        else:
            # If the cancellation is propagated normally and the awaited (sub-) tasks exited,
            # consider it as a successful cleanup.
            if logger is not None and (not quiet or pending or iterations > 1):
                are = 'are' if not pending else 'are not'
                why = 'cancelling normally' if cancelled else 'finishing normally'
                logger.debug(f"{captitle} tasks {are} stopped: {why}; tasks left: {pending!r}")
            done_ever |= done_now

    return done_ever, pending


async def reraise(
        tasks: Collection[Task],
) -> None:
    """
    Re-raise errors from tasks, if any. Do nothing if all tasks have succeeded.
    """
    for task in tasks:
        try:
            task.result()  # can raise the regular (non-cancellation) exceptions.
        except asyncio.CancelledError:
            pass


async def all_tasks(
        *,
        ignored: Collection[Task] = frozenset(),
) -> Collection[Task]:
    """
    Return all tasks in the current event loop.

    Equivalent to :func:`asyncio.all_tasks`, but with an exlcusion list.
    The exclusion list is used to exclude the tasks that existed at a point
    in time in the past, to only get the tasks that appeared since then.
    """
    current_task = asyncio.current_task()
    return {task for task in asyncio.all_tasks()
            if task is not current_task and task not in ignored}



from kopf.reactor.queueing import get_uid


def test_uid_is_used_if_present():
    raw_event = {'type': ..., 'object': {'metadata': {'uid': '123'}}}
    uid = get_uid(raw_event)

    assert isinstance(uid, str)
    assert uid == '123'


def test_uid_is_simulated_if_absent():
    raw_event = {'type': ...,
                 'object': {
                     'apiVersion': 'group/v1',
                     'kind': 'Kind1',
                     'metadata': {
                         'name': 'name1',
                         'namespace': 'namespace1',
                         'creationTimestamp': 'created1',
                     }}}
    uid = get_uid(raw_event)

    # The exact order is irrelevant.
    assert isinstance(uid, str)
    assert 'created1' in uid
    assert 'name1' in uid
    assert 'namespace' in uid
    assert 'Kind1' in uid
    assert 'group/v1' in uid



"""
Rudimentary piggybacking on the known K8s API clients for authentication.

Kopf is not a client library, and avoids bringing too much logic
for proper authentication, especially all the complex auth-providers.

Instead, it uses the existing clients, triggers the (re-)authentication
in them, and extracts the basic credentials for its own use.

.. seealso::
    :mod:`credentials` and :func:`authentication`.
"""
import logging
from typing import Any, Optional, Sequence, Union

from kopf.structs import credentials

# Keep as constants to make them patchable. Higher priority is more preferred.
PRIORITY_OF_CLIENT: int = 10
PRIORITY_OF_PYKUBE: int = 20


# We keep the official client library auto-login only because it was
# an implied behavior before switching to pykube -- to keep it so (implied).
def login_via_client(
        *args: Any,
        logger: Union[logging.Logger, logging.LoggerAdapter],
        **kwargs: Any,
) -> Optional[credentials.ConnectionInfo]:

    try:
        import kubernetes.config
    except ImportError:
        return None

    try:
        kubernetes.config.load_incluster_config()  # cluster env vars
        logger.debug("Client is configured in cluster with service account.")
    except kubernetes.config.ConfigException as e1:
        try:
            kubernetes.config.load_kube_config()  # developer's config files
            logger.debug("Client is configured via kubeconfig file.")
        except kubernetes.config.ConfigException as e2:
            raise credentials.LoginError(f"Cannot authenticate client neither in-cluster, nor via kubeconfig.")

    # We do not even try to understand how it works and why. Just load it, and extract the results.
    # For kubernetes client >= 12.0.0 use the new 'get_default_copy' method
    if callable(getattr(kubernetes.client.Configuration, 'get_default_copy', None)):
        config = kubernetes.client.Configuration.get_default_copy()
    else:
        config = kubernetes.client.Configuration()

    # For auth-providers, this method is monkey-patched with the auth-provider's one.
    # We need the actual auth-provider's token, so we call it instead of accessing api_key.
    # Other keys (token, tokenFile) also end up being retrieved via this method.
    header: Optional[str] = config.get_api_key_with_prefix('authorization')
    parts: Sequence[str] = header.split(' ', 1) if header else []
    scheme, token = ((None, None) if len(parts) == 0 else
                     (None, parts[0]) if len(parts) == 1 else
                     (parts[0], parts[1]))  # RFC-7235, Appendix C.

    # Interpret the config object for our own minimalistic credentials.
    # Note: kubernetes client has no concept of a "current" context's namespace.
    return credentials.ConnectionInfo(
        server=config.host,
        ca_path=config.ssl_ca_cert,  # can be a temporary file
        insecure=not config.verify_ssl,
        username=config.username or None,  # an empty string when not defined
        password=config.password or None,  # an empty string when not defined
        scheme=scheme,
        token=token,
        certificate_path=config.cert_file,  # can be a temporary file
        private_key_path=config.key_file,  # can be a temporary file
        priority=PRIORITY_OF_CLIENT,
    )


# Pykube login is mandatory. If it fails, the framework will not run at all.
def login_via_pykube(
        *args: Any,
        logger: Union[logging.Logger, logging.LoggerAdapter],
        **kwargs: Any,
) -> Optional[credentials.ConnectionInfo]:

    try:
        import pykube
    except ImportError:
        return None

    # Read the pykube config either way for later interpretation.
    config: pykube.KubeConfig
    try:
        config = pykube.KubeConfig.from_service_account()
        logger.debug("Pykube is configured in cluster with service account.")
    except FileNotFoundError:
        try:
            config = pykube.KubeConfig.from_file()
            logger.debug("Pykube is configured via kubeconfig file.")
        except (pykube.PyKubeError, FileNotFoundError):
            raise credentials.LoginError(f"Cannot authenticate pykube "
                                         f"neither in-cluster, nor via kubeconfig.")

    # We don't know how this token will be retrieved, we just get it afterwards.
    provider_token = None
    if config.user.get('auth-provider'):
        api = pykube.HTTPClient(config)
        api.get(version='', base='/')  # ignore the response status
        provider_token = config.user.get('auth-provider', {}).get('config', {}).get('access-token')

    # Interpret the config object for our own minimalistic credentials.
    ca: Optional[pykube.config.BytesOrFile] = config.cluster.get('certificate-authority')
    cert: Optional[pykube.config.BytesOrFile] = config.user.get('client-certificate')
    pkey: Optional[pykube.config.BytesOrFile] = config.user.get('client-key')
    return credentials.ConnectionInfo(
        server=config.cluster.get('server'),
        ca_path=ca.filename() if ca else None,  # can be a temporary file
        insecure=config.cluster.get('insecure-skip-tls-verify'),
        username=config.user.get('username'),
        password=config.user.get('password'),
        token=config.user.get('token') or provider_token,
        certificate_path=cert.filename() if cert else None,  # can be a temporary file
        private_key_path=pkey.filename() if pkey else None,  # can be a temporary file
        default_namespace=config.namespace,
        priority=PRIORITY_OF_PYKUBE,
    )



import os.path
import time

import kopf.testing
import pykube


def test_pods_reacted():

    example_py = os.path.join(os.path.dirname(__file__), 'example.py')
    with kopf.testing.KopfRunner(
        ['run', '--all-namespaces', '--standalone', '--verbose', example_py],
        timeout=60,
    ) as runner:
        name = _create_pod()
        time.sleep(5)  # give it some time to react
        _delete_pod(name)
        time.sleep(1)  # give it some time to react

    assert runner.exception is None
    assert runner.exit_code == 0

    assert f'[default/{name}] Creation is in progress:' in runner.stdout
    assert f'[default/{name}] === Pod killing happens in 30s.' in runner.stdout
    assert f'[default/{name}] Deletion is in progress:' in runner.stdout
    assert f'[default/{name}] === Pod killing is cancelled!' in runner.stdout


def _create_pod():
    api = pykube.HTTPClient(pykube.KubeConfig.from_file())
    with api.session:
        pod = pykube.Pod(api, {
            'apiVersion': 'v1',
            'kind': 'Pod',
            'metadata': {'generateName': 'kopf-pod-', 'namespace': 'default'},
            'spec': {
                'containers': [{
                    'name': 'the-only-one',
                    'image': 'busybox',
                    'command': ["sh", "-x", "-c", "sleep 1"],
                }]},
        })
        pod.create()
        return pod.name


def _delete_pod(name):
    api = pykube.HTTPClient(pykube.KubeConfig.from_file())
    with api.session:
        pod = pykube.Pod.objects(api, namespace='default').get_by_name(name)
        pod.delete()



import pytest


class CustomIterable:
    def __init__(self, objs):
        self._objs = objs

    def __iter__(self):
        for obj in self._objs:
            yield obj


@pytest.fixture(params=[list, tuple, CustomIterable],
                ids=['list', 'tuple', 'custom'])
def multicls(request):
    return request.param



import collections.abc

import pytest

from kopf.reactor.causation import ResourceChangingCause, ResourceWatchingCause
from kopf.structs.handlers import Activity


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn():
    pass


def test_generic_registry_via_iter(
        generic_registry_cls, cause_factory):

    cause = cause_factory(generic_registry_cls)
    registry = generic_registry_cls()
    iterator = registry.iter_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_generic_registry_via_list(
        generic_registry_cls, cause_factory):

    cause = cause_factory(generic_registry_cls)
    registry = generic_registry_cls()
    handlers = registry.get_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


@pytest.mark.parametrize('activity', list(Activity))
def test_operator_registry_with_activity_via_iter(
        operator_registry_cls, activity):

    registry = operator_registry_cls()
    iterator = registry._activities.iter_handlers(activity=activity)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_operator_registry_with_resource_watching_via_iter(
        operator_registry_cls, cause_factory):

    cause = cause_factory(ResourceWatchingCause)
    registry = operator_registry_cls()
    iterator = registry._resource_watching.iter_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


def test_operator_registry_with_resource_changing_via_iter(
        operator_registry_cls, cause_factory):

    cause = cause_factory(ResourceChangingCause)
    registry = operator_registry_cls()
    iterator = registry._resource_changing.iter_handlers(cause)

    assert isinstance(iterator, collections.abc.Iterator)
    assert not isinstance(iterator, collections.abc.Collection)
    assert not isinstance(iterator, collections.abc.Container)
    assert not isinstance(iterator, (list, tuple))

    handlers = list(iterator)
    assert not handlers


@pytest.mark.parametrize('activity', list(Activity))
def test_operator_registry_with_activity_via_list(
        operator_registry_cls, activity):

    registry = operator_registry_cls()
    handlers = registry._activities.get_handlers(activity=activity)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_operator_registry_with_resource_watching_via_list(
        operator_registry_cls, cause_factory):

    cause = cause_factory(ResourceWatchingCause)
    registry = operator_registry_cls()
    handlers = registry._resource_watching.get_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers


def test_operator_registry_with_resource_changing_via_list(
        operator_registry_cls, cause_factory):

    cause = cause_factory(ResourceChangingCause)
    registry = operator_registry_cls()
    handlers = registry._resource_changing.get_handlers(cause)

    assert isinstance(handlers, collections.abc.Iterable)
    assert isinstance(handlers, collections.abc.Container)
    assert isinstance(handlers, collections.abc.Collection)
    assert not handlers



import asynctest

from kopf.clients.watching import _iter_jsonlines


async def test_empty_content():
    async def iter_chunked(n: int):
        if False:  # to make this function a generator
            yield b''

    content = asynctest.Mock(iter_chunked=iter_chunked)
    lines = []
    async for line in _iter_jsonlines(content):
        lines.append(line)

    assert lines == []


async def test_empty_chunk():
    async def iter_chunked(n: int):
        yield b''

    content = asynctest.Mock(iter_chunked=iter_chunked)
    lines = []
    async for line in _iter_jsonlines(content):
        lines.append(line)

    assert lines == []


async def test_one_chunk_one_line():
    async def iter_chunked(n: int):
        yield b'hello'

    content = asynctest.Mock(iter_chunked=iter_chunked)
    lines = []
    async for line in _iter_jsonlines(content):
        lines.append(line)

    assert lines == [b'hello']


async def test_one_chunk_two_lines():
    async def iter_chunked(n: int):
        yield b'hello\nworld'

    content = asynctest.Mock(iter_chunked=iter_chunked)
    lines = []
    async for line in _iter_jsonlines(content):
        lines.append(line)

    assert lines == [b'hello', b'world']


async def test_one_chunk_empty_lines():
    async def iter_chunked(n: int):
        yield b'\n\nhello\n\nworld\n\n'

    content = asynctest.Mock(iter_chunked=iter_chunked)
    lines = []
    async for line in _iter_jsonlines(content):
        lines.append(line)

    assert lines == [b'hello', b'world']


async def test_few_chunks_split():
    async def iter_chunked(n: int):
        yield b'\n\nhell'
        yield b'o\n\nwor'
        yield b'ld\n\n'

    content = asynctest.Mock(iter_chunked=iter_chunked)
    lines = []
    async for line in _iter_jsonlines(content):
        lines.append(line)

    assert lines == [b'hello', b'world']



import pytest

from kopf.structs.references import EVERYTHING, Resource, Selector


@pytest.fixture()
def resource():
    return Resource(
        group='group1', version='version1', preferred=True,
        plural='plural1', singular='singular1', kind='kind1',
        shortcuts=['shortcut1', 'shortcut2'],
        categories=['category1', 'category2'],
    )


@pytest.fixture()
def v1_resource():
    return Resource(
        group='', version='v1', preferred=True,
        plural='plural1', singular='singular1', kind='kind1',
        shortcuts=['shortcut1', 'shortcut2'],
        categories=['category1', 'category2'],
    )


@pytest.mark.parametrize('group, version', [
    (None, None),
    ('group1', None),
    (None, 'version1'),
    ('group1', 'version1'),
])
def test_when_matches_everything(resource, group, version):
    selector = Selector(EVERYTHING, group=group, version=version)
    matches = selector.check(resource)
    assert matches


@pytest.mark.parametrize('kwarg, kwval', [
    ('kind', 'kind1'),
    ('plural', 'plural1'),
    ('singular', 'singular1'),
    ('shortcut', 'shortcut1'),
    ('shortcut', 'shortcut2'),
    ('category', 'category1'),
    ('category', 'category2'),
    ('any_name', 'kind1'),
    ('any_name', 'plural1'),
    ('any_name', 'singular1'),
    ('any_name', 'shortcut1'),
    ('any_name', 'shortcut2'),
])
@pytest.mark.parametrize('group, version', [
    (None, None),
    ('group1', None),
    (None, 'version1'),
    ('group1', 'version1'),
])
def test_when_matches_names(resource, kwarg, kwval, group, version):
    selector = Selector(group=group, version=version, **{kwarg: kwval})
    matches = selector.check(resource)
    assert matches


@pytest.mark.parametrize('kwarg, kwval', [
    ('kind', 'kind1'),
    ('plural', 'plural1'),
    ('singular', 'singular1'),
    ('shortcut', 'shortcut1'),
    ('shortcut', 'shortcut2'),
    ('category', 'category1'),
    ('category', 'category2'),
    ('any_name', 'kind1'),
    ('any_name', 'plural1'),
    ('any_name', 'singular1'),
    ('any_name', 'shortcut1'),
    ('any_name', 'shortcut2'),
])
@pytest.mark.parametrize('group, version', [
    ('group9', None),
    (None, 'version9'),
    ('group1', 'version9'),
    ('group9', 'version1'),
    ('group9', 'version9'),
])
def test_when_groupversion_mismatch_but_names_do_match(resource, kwarg, kwval, group, version):
    selector = Selector(group=group, version=version, **{kwarg: kwval})
    matches = selector.check(resource)
    assert not matches


@pytest.mark.parametrize('kwarg, kwval', [
    ('kind', 'kind9'),
    ('plural', 'plural9'),
    ('singular', 'singular9'),
    ('shortcut', 'shortcut9'),
    ('category', 'category9'),
    ('any_name', 'category9'),
    ('any_name', 'category1'),  # categories are not used with any_name, must be explicit.
    ('any_name', 'category2'),  # categories are not used with any_name, must be explicit.
])
@pytest.mark.parametrize('group, version', [
    (None, None),
    ('group1', None),
    (None, 'version1'),
    ('group1', 'version1'),
])
def test_when_groupversion_do_match_but_names_mismatch(resource, kwarg, kwval, group, version):
    selector = Selector(group=group, version=version, **{kwarg: kwval})
    matches = selector.check(resource)
    assert not matches


def test_catchall_versions_are_ignored_for_nonpreferred_resources():
    resource = Resource(
        group='group1', version='version1', preferred=False,
        plural='plural1', singular='singular1', kind='kind1',
        shortcuts=['shortcut1', 'shortcut2'],
        categories=['category1', 'category2'],
    )
    selector = Selector(EVERYTHING)
    matches = selector.check(resource)
    assert not matches


@pytest.mark.parametrize('selector_args', [
    pytest.param(['events'], id='only-name'),
    pytest.param(['v1', 'events'], id='with-version'),
    pytest.param(['', 'v1', 'events'], id='with-groupversion'),
])
def test_events_are_matched_when_explicitly_named(selector_args):
    resource = Resource('', 'v1', 'events')
    selector = Selector(*selector_args)
    matches = selector.check(resource)
    assert matches


@pytest.mark.parametrize('selector_args', [
    pytest.param([EVERYTHING], id='only-marker'),
    pytest.param(['v1', EVERYTHING], id='with-core-version'),
    pytest.param(['', 'v1', EVERYTHING], id='with-core-groupversion'),
    pytest.param(['events.k8s.io', EVERYTHING], id='with-k8sio-group'),
    pytest.param(['events.k8s.io', 'v1beta1', EVERYTHING], id='with-k8sio-groupversion'),
])
@pytest.mark.parametrize('resource_kwargs', [
    pytest.param(dict(group='', version='v1'), id='core-v1'),
    pytest.param(dict(group='events.k8s.io', version='v1'), id='k8sio-v1'),
    pytest.param(dict(group='events.k8s.io', version='v1beta1'), id='k8sio-v1beta1'),
])
def test_events_are_excluded_from_everything(resource_kwargs, selector_args):
    resource = Resource(**resource_kwargs, plural='events')
    selector = Selector(*selector_args)
    matches = selector.check(resource)
    assert not matches


@pytest.mark.parametrize('kwarg, kwval', [
    ('kind', 'kind1'),
    ('plural', 'plural1'),
    ('singular', 'singular1'),
    ('shortcut', 'shortcut1'),
    ('shortcut', 'shortcut2'),
    ('any_name', 'kind1'),
    ('any_name', 'plural1'),
    ('any_name', 'singular1'),
    ('any_name', 'shortcut1'),
    ('any_name', 'shortcut2'),
])
def test_selection_of_specific_resources(resource, kwarg, kwval):
    selector = Selector(**{kwarg: kwval})
    selected = selector.select([resource])
    assert selector.is_specific  # prerequisite
    assert selected == {resource}


@pytest.mark.parametrize('kwarg, kwval', [
    ('category', 'category1'),
    ('category', 'category2'),
    ('any_name', EVERYTHING),
])
def test_selection_of_nonspecific_resources(resource, kwarg, kwval):
    selector = Selector(**{kwarg: kwval})
    selected = selector.select([resource])
    assert not selector.is_specific  # prerequisite
    assert selected == {resource}


@pytest.mark.parametrize('kwarg, kwval', [
    ('kind', 'kind1'),
    ('plural', 'plural1'),
    ('singular', 'singular1'),
    ('shortcut', 'shortcut1'),
    ('shortcut', 'shortcut2'),
    ('any_name', 'kind1'),
    ('any_name', 'plural1'),
    ('any_name', 'singular1'),
    ('any_name', 'shortcut1'),
    ('any_name', 'shortcut2'),
])
def test_precedence_of_corev1_over_others_when_specific(resource, v1_resource, kwarg, kwval):
    selector = Selector(**{kwarg: kwval})
    selected = selector.select([resource, v1_resource])
    assert selector.is_specific  # prerequisite
    assert selected == {v1_resource}


@pytest.mark.parametrize('kwarg, kwval', [
    ('category', 'category1'),
    ('category', 'category2'),
    ('any_name', EVERYTHING),
])
def test_precedence_of_corev1_same_as_others_when_nonspecific(resource, v1_resource, kwarg, kwval):
    selector = Selector(**{kwarg: kwval})
    selected = selector.select([resource, v1_resource])
    assert not selector.is_specific  # prerequisite
    assert selected == {resource, v1_resource}



import pytest

import kopf
from kopf.reactor.observation import revise_resources
from kopf.structs.references import EVERYTHING, Insights, Resource

VERBS = ['list', 'watch', 'patch']


@pytest.fixture(params=[
    kopf.on.event, kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def handlers(request, registry):
    @request.param('group1', 'version1', 'plural1')
    @request.param('group2', 'version2', 'plural2')
    def fn(**_): ...


@pytest.mark.usefixtures('handlers')
def test_initial_population(registry):
    r1 = Resource(group='group1', version='version1', plural='plural1', verbs=VERBS)
    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1])
    assert insights.resources == {r1}


@pytest.mark.usefixtures('handlers')
def test_replacing_all_insights(registry):
    r1 = Resource(group='group1', version='version1', plural='plural1', verbs=VERBS)
    r2 = Resource(group='group2', version='version2', plural='plural2', verbs=VERBS)
    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1])
    revise_resources(registry=registry, insights=insights, group=None, resources=[r2])
    assert insights.resources == {r2}


@pytest.mark.usefixtures('handlers')
def test_replacing_an_existing_group(registry):
    r1 = Resource(group='group1', version='version1', plural='plural1', verbs=VERBS)
    r2 = Resource(group='group2', version='version2', plural='plural2', verbs=VERBS)
    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1])
    revise_resources(registry=registry, insights=insights, group='group1', resources=[r2])
    assert insights.resources == {r2}


@pytest.mark.usefixtures('handlers')
def test_replacing_a_new_group(registry):
    r1 = Resource(group='group1', version='version1', plural='plural1', verbs=VERBS)
    r2 = Resource(group='group2', version='version2', plural='plural2', verbs=VERBS)
    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1])
    revise_resources(registry=registry, insights=insights, group='group2', resources=[r2])
    assert insights.resources == {r1, r2}


@pytest.mark.parametrize('decorator', [
    kopf.on.event, kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def test_ambiguity_in_specific_selectors(registry, decorator, caplog, assert_logs):
    r1 = Resource(group='g1', version='v1', plural='plural', verbs=VERBS)
    r2 = Resource(group='g2', version='v2', plural='plural', verbs=VERBS)

    @decorator(plural='plural')
    def fn(**_): ...

    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1, r2])
    assert not insights.resources
    assert_logs([r"Ambiguous resources will not be served"])


@pytest.mark.parametrize('decorator', [
    kopf.on.event, kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def test_corev1_overrides_ambuigity(registry, decorator, caplog, assert_logs):
    r1 = Resource(group='', version='v1', plural='pods', verbs=VERBS)
    r2 = Resource(group='metrics.k8s.io', version='v1', plural='pods', verbs=VERBS)

    @decorator(plural='pods')
    def fn(**_): ...

    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1, r2])
    assert insights.resources == {r1}
    assert_logs([], prohibited=[r"Ambiguous resources will not be served"])


@pytest.mark.parametrize('decorator', [
    kopf.on.event, kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def test_no_ambiguity_in_generic_selector(registry, decorator, caplog, assert_logs):
    r1 = Resource(group='g1', version='v1', plural='plural', verbs=VERBS)
    r2 = Resource(group='g2', version='v2', plural='plural', verbs=VERBS)

    @decorator(EVERYTHING)
    def fn(**_): ...

    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1, r2])
    assert insights.resources == {r1, r2}
    assert_logs([], prohibited=[r"Ambiguous resources will not be served"])


@pytest.mark.parametrize('decorator', [
    kopf.on.event, kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def test_selectors_with_no_resources(registry, decorator, caplog, assert_logs):
    r1 = Resource(group='group1', version='version1', plural='plural1', verbs=VERBS)
    r2 = Resource(group='group2', version='version2', plural='plural2', verbs=VERBS)

    @decorator(plural='plural3')
    def fn(**_): ...

    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1, r2])
    assert not insights.resources
    assert_logs([r"Unresolved resources cannot be served"])


@pytest.mark.parametrize('decorator', [
    kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def test_nonwatchable_excluded(registry, decorator, caplog, assert_logs):
    r1 = Resource(group='group1', version='version1', plural='plural1', verbs=[])

    @decorator('group1', 'version1', 'plural1')
    def fn(**_): ...

    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1])
    assert not insights.resources
    assert_logs([r"Non-watchable resources will not be served: {plural1.version1.group1}"])


@pytest.mark.parametrize('decorator', [
    kopf.daemon, kopf.timer,
    kopf.on.resume, kopf.on.create, kopf.on.update, kopf.on.delete,
])
def test_nonpatchable_excluded(registry, decorator, caplog, assert_logs):
    r1 = Resource(group='group1', version='version1', plural='plural1', verbs=['watch', 'list'])

    @decorator('group1', 'version1', 'plural1')  # because it patches!
    def fn(**_): ...

    insights = Insights()
    revise_resources(registry=registry, insights=insights, group=None, resources=[r1])
    assert not insights.resources
    assert_logs([r"Non-patchable resources will not be served: {plural1.version1.group1}"])



"""
All the functions to manipulate the object finalization and deletion.

Finalizers are used to block the actual deletion until the finalizers
are removed, meaning that the operator has done all its duties
to "release" the object (e.g. cleanups; delete-handlers in our case).
"""
from kopf.structs import bodies, patches


def is_deletion_ongoing(
        body: bodies.Body,
) -> bool:
    return body.get('metadata', {}).get('deletionTimestamp', None) is not None


def is_deletion_blocked(
        body: bodies.Body,
        finalizer: str,
) -> bool:
    finalizers = body.get('metadata', {}).get('finalizers', [])
    return finalizer in finalizers


def block_deletion(
        *,
        body: bodies.Body,
        patch: patches.Patch,
        finalizer: str,
) -> None:
    if not is_deletion_blocked(body=body, finalizer=finalizer):
        finalizers = body.get('metadata', {}).get('finalizers', [])
        patch.setdefault('metadata', {}).setdefault('finalizers', list(finalizers))
        patch['metadata']['finalizers'].append(finalizer)


def allow_deletion(
        *,
        body: bodies.Body,
        patch: patches.Patch,
        finalizer: str,
) -> None:
    if is_deletion_blocked(body=body, finalizer=finalizer):
        finalizers = body.get('metadata', {}).get('finalizers', [])
        patch.setdefault('metadata', {}).setdefault('finalizers', list(finalizers))
        if finalizer in patch['metadata']['finalizers']:
            patch['metadata']['finalizers'].remove(finalizer)
